"""
This module is a factory method what allows to import various objects, such as algorithms, crossover, mutation.

The definitions for each object are purposely defined as a list and not as a dictionary to keep an order for the documentation.


"""
import re

from pymoo.configuration import Configuration
from pymoo.problems.many import *
from pymoo.problems.multi import *
from pymoo.problems.single import *


# =========================================================================================================
# Generic
# =========================================================================================================



def get_from_list(l, name, args, kwargs):
    i = None

    for k, e in enumerate(l):
        if e[0] == name:
            i = k
            break

    if i is None:
        for k, e in enumerate(l):
            if re.match(e[0], name):
                i = k
                break

    if i is not None:

        if len(l[i]) == 2:
            name, clazz = l[i]

        elif len(l[i]) == 3:
            name, clazz, default_kwargs = l[i]

            # overwrite the default if provided
            for key, val in kwargs.items():
                default_kwargs[key] = val
            kwargs = default_kwargs

        return clazz(*args, **kwargs)
    else:
        raise Exception("Object '%s' for not found in %s" % (name, [e[0] for e in l]))


# =========================================================================================================
# Algorithms
# =========================================================================================================

def get_algorithm_options():
    from pymoo.algorithms.moead import MOEAD
    from pymoo.algorithms.nsga2 import NSGA2
    from pymoo.algorithms.nsga3 import NSGA3
    from pymoo.algorithms.rnsga2 import RNSGA2
    from pymoo.algorithms.rnsga3 import RNSGA3
    from pymoo.algorithms.so_de import DE
    from pymoo.algorithms.so_genetic_algorithm import GA
    from pymoo.algorithms.unsga3 import UNSGA3
    from pymoo.algorithms.so_nelder_mead import NelderMead
    from pymoo.algorithms.so_cmaes import CMAES
    from pymoo.algorithms.so_brkga import BRKGA
    from pymoo.algorithms.so_pattern_search import PatternSearch

    ALGORITHMS = [
        ("ga", GA),
        ("brkga", BRKGA),
        ("de", DE),
        ("nelder-mead", NelderMead),
        ("pattern-search", PatternSearch),
        ("cmaes", CMAES),
        ("nsga2", NSGA2),
        ("rnsga2", RNSGA2),
        ("nsga3", NSGA3),
        ("unsga3", UNSGA3),
        ("rnsga3", RNSGA3),
        ("moead", MOEAD),
    ]

    return ALGORITHMS


def get_algorithm(name, *args, d={}, **kwargs):
    return get_from_list(get_algorithm_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Sampling
# =========================================================================================================

def get_sampling_options():
    from pymoo.operators.sampling.latin_hypercube_sampling import LatinHypercubeSampling
    from pymoo.operators.sampling.random_sampling import FloatRandomSampling
    from pymoo.operators.integer_from_float_operator import IntegerFromFloatSampling
    from pymoo.operators.sampling.random_sampling import BinaryRandomSampling

    SAMPLING = [
        ("real_random", FloatRandomSampling),
        ("real_lhs", LatinHypercubeSampling),
        ("bin_random", BinaryRandomSampling),
        ("int_random", IntegerFromFloatSampling, {'clazz': FloatRandomSampling}),
        ("int_lhs", IntegerFromFloatSampling, {'clazz': LatinHypercubeSampling})
    ]

    return SAMPLING


def get_sampling(name, *args, d={}, **kwargs):
    return get_from_list(get_sampling_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Selection
# =========================================================================================================

def get_selection_options():
    from pymoo.operators.selection.random_selection import RandomSelection
    from pymoo.operators.selection.tournament_selection import TournamentSelection

    SELECTION = [
        ("random", RandomSelection),
        ("tournament", TournamentSelection)
    ]

    return SELECTION


def get_selection(name, *args, d={}, **kwargs):
    return get_from_list(get_selection_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Crossover
# =========================================================================================================

def get_crossover_options():
    from pymoo.operators.crossover.differental_evolution_crossover import DifferentialEvolutionCrossover
    from pymoo.operators.crossover.exponential_crossover import ExponentialCrossover
    from pymoo.operators.crossover.half_uniform_crossover import HalfUniformCrossover
    from pymoo.operators.crossover.point_crossover import PointCrossover
    from pymoo.operators.crossover.simulated_binary_crossover import SimulatedBinaryCrossover
    from pymoo.operators.crossover.uniform_crossover import UniformCrossover
    from pymoo.operators.integer_from_float_operator import IntegerFromFloatCrossover

    CROSSOVER = [
        ("real_sbx", SimulatedBinaryCrossover, dict(prob=0.9, eta=30)),
        ("int_sbx", IntegerFromFloatCrossover, dict(clazz=SimulatedBinaryCrossover, prob=0.9, eta=30)),
        ("real_de", DifferentialEvolutionCrossover),
        ("(real|bin|int)_ux", UniformCrossover),
        ("(bin|int)_hux", HalfUniformCrossover),
        ("(real|bin|int)_exp", ExponentialCrossover),
        ("(real|bin|int)_one_point", PointCrossover, {'n_points': 1}),
        ("(real|bin|int)_two_point", PointCrossover, {'n_points': 2}),
        ("(real|bin|int)_k_point", PointCrossover)
    ]

    return CROSSOVER


def get_crossover(name, *args, d={}, **kwargs):
    return get_from_list(get_crossover_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Mutation
# =========================================================================================================

def get_mutation_options():
    from pymoo.operators.mutation.no_mutation import NoMutation
    from pymoo.operators.mutation.bitflip_mutation import BinaryBitflipMutation
    from pymoo.operators.mutation.polynomial_mutation import PolynomialMutation
    from pymoo.operators.integer_from_float_operator import IntegerFromFloatMutation

    MUTATION = [
        ("none", NoMutation, {}),
        ("real_pm", PolynomialMutation, dict(eta=20)),
        ("int_pm", IntegerFromFloatMutation, dict(clazz=PolynomialMutation, eta=20)),
        ("bin_bitflip", BinaryBitflipMutation)
    ]

    return MUTATION


def get_mutation(name, *args, d={}, **kwargs):
    return get_from_list(get_mutation_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Termination
# =========================================================================================================

def get_termination_options():
    from pymoo.util.termination.max_eval import MaximumFunctionCallTermination
    from pymoo.util.termination.max_gen import MaximumGenerationTermination
    from pymoo.util.termination.max_time import TimeBasedTermination
    from pymoo.util.termination.min_igd import IGDTermination
    from pymoo.util.termination.x_tol import DesignSpaceToleranceTermination
    from pymoo.util.termination.f_tol import MultiObjectiveSpaceToleranceTermination
    from pymoo.util.termination.f_tol_single import SingleObjectiveSpaceToleranceTermination
    from pymoo.util.termination.default import MultiObjectiveDefaultTermination, SingleObjectiveDefaultTermination

    TERMINATION = [
        ("n_eval", MaximumFunctionCallTermination),
        ("(n_gen|n_iter)", MaximumGenerationTermination),
        ("igd", IGDTermination),
        ("time", TimeBasedTermination),
        ("(x_tol|xtol)", DesignSpaceToleranceTermination),
        ("(f_tol$|ftol$)", MultiObjectiveSpaceToleranceTermination),
        ("(f_tol_s|ftol_s)", SingleObjectiveSpaceToleranceTermination),
        ("(default$|default_multi)", MultiObjectiveDefaultTermination),
        ("default_single$", SingleObjectiveDefaultTermination)
    ]

    return TERMINATION


def get_termination(name, *args, d={}, **kwargs):
    return get_from_list(get_termination_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Problems
# =========================================================================================================

def get_problem_options():
    PROBLEM = [
        ('ackley', Ackley),
        ('bnh', BNH),
        ('carside', Carside),
        ('ctp1', CTP1),
        ('ctp2', CTP2),
        ('ctp3', CTP3),
        ('ctp4', CTP4),
        ('ctp5', CTP5),
        ('ctp6', CTP6),
        ('ctp7', CTP7),
        ('ctp8', CTP8),
        ('dtlz1^-1', InvertedDTLZ1),
        ('dtlz1', DTLZ1),
        ('dtlz2', DTLZ2),
        ('dtlz3', DTLZ3),
        ('dtlz4', DTLZ4),
        ('dtlz5', DTLZ5),
        ('dtlz6', DTLZ6),
        ('dtlz7', DTLZ7),
        ('convex_dtlz2', ConvexDTLZ2),
        ('convex_dtlz4', ConvexDTLZ4),
        ('sdtlz1', ScaledDTLZ1),
        ('c1dtlz1', C1DTLZ1),
        ('c1dtlz3', C1DTLZ3),
        ('c2dtlz2', C2DTLZ2),
        ('c3dtlz4', C3DTLZ4),
        ('cantilevered_beam', CantileveredBeam),
        ('griewank', Griewank),
        ('himmelblau', Himmelblau),
        ('knp', Knapsack),
        ('kursawe', Kursawe),
        ('osy', OSY),
        ('pressure_vessel', PressureVessel),
        ('rastrigin', Rastrigin),
        ('rosenbrock', Rosenbrock),
        ('schwefel', Schwefel),
        ('sphere', Sphere),
        ('tnk', TNK),
        ('truss2d', Truss2D),
        ('welded_beam', WeldedBeam),
        ('zakharov', Zakharov),
        ('zdt1', ZDT1),
        ('zdt2', ZDT2),
        ('zdt3', ZDT3),
        ('zdt4', ZDT4),
        ('zdt5', ZDT5),
        ('zdt6', ZDT6),
        ('g01', G1),
        ('g02', G2),
        ('g03', G3),
        ('g04', G4),
        ('g05', G5),
        ('g06', G6),
        ('g07', G7),
        ('g08', G8),
        ('g09', G9),
        ('g10', G10),
        ('wfg1', WFG1),
        ('wfg2', WFG2),
        ('wfg3', WFG3),
        ('wfg4', WFG4),
        ('wfg5', WFG5),
        ('wfg6', WFG6),
        ('wfg7', WFG7),
        ('wfg8', WFG8),
        ('wfg9', WFG9)
    ]

    return PROBLEM


def get_problem(name, *args, d={}, **kwargs):
    if name.startswith("go-"):
        from pymoo.vendor.global_opt import get_global_optimization_problem_options
        return get_from_list(get_global_optimization_problem_options(), name.lower(), args, {**d, **kwargs})
    else:
        return get_from_list(get_problem_options(), name.lower(), args, {**d, **kwargs})


# =========================================================================================================
# Weights
# =========================================================================================================

def get_reference_direction_options():
    from pymoo.util.reference_direction import UniformReferenceDirectionFactory
    from pymoo.util.reference_direction import MultiLayerReferenceDirectionFactory
    from pymoo.util.ref_dirs.reduction import ReductionBasedReferenceDirectionFactory
    from pymoo.util.ref_dirs.energy import RieszEnergyReferenceDirectionFactory
    from pymoo.util.ref_dirs.energy_layer import LayerwiseRieszEnergyReferenceDirectionFactory

    REFERENCE_DIRECTIONS = [
        ("(das-dennis|uniform)", UniformReferenceDirectionFactory),
        ("multi-layer", MultiLayerReferenceDirectionFactory),
        ("(energy|riesz)", RieszEnergyReferenceDirectionFactory),
        ("(layer-energy|layer-riesz)", LayerwiseRieszEnergyReferenceDirectionFactory),
        ("red", ReductionBasedReferenceDirectionFactory)
    ]

    return REFERENCE_DIRECTIONS


def get_reference_directions(name, *args, d={}, **kwargs):
    return get_from_list(get_reference_direction_options(), name, args, {**d, **kwargs}).do()


# =========================================================================================================
# Visualization
# =========================================================================================================

def get_visualization_options():
    from pymoo.visualization.pcp import PCP
    from pymoo.visualization.petal import Petal
    from pymoo.visualization.radar import Radar
    from pymoo.visualization.radviz import Radviz
    from pymoo.visualization.scatter import Scatter
    from pymoo.visualization.star_coordinate import StarCoordinate
    from pymoo.visualization.heatmap import Heatmap
    from pymoo.visualization.fitness_landscape import FitnessLandscape

    VISUALIZATION = [
        ("scatter", Scatter),
        ("heatmap", Heatmap),
        ("pcp", PCP),
        ("petal", Petal),
        ("radar", Radar),
        ("radviz", Radviz),
        ("star", StarCoordinate),
        ("fitness-landscape", FitnessLandscape)
    ]

    return VISUALIZATION


def get_visualization(name, *args, d={}, **kwargs):
    return get_from_list(get_visualization_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Performance Indicator
# =========================================================================================================


def get_performance_indicator_options():
    from pymoo.performance_indicator.gd import GD
    from pymoo.performance_indicator.gd_plus import GDPlus
    from pymoo.performance_indicator.igd import IGD
    from pymoo.performance_indicator.igd_plus import IGDPlus
    from pymoo.performance_indicator.hv import Hypervolume
    from pymoo.performance_indicator.rmetric import RMetric

    PERFORMANCE_INDICATOR = [
        ("gd", GD),
        ("gd+", GDPlus),
        ("igd", IGD),
        ("igd+", IGDPlus),
        ("hv", Hypervolume),
        ("rmetric", RMetric)
    ]
    return PERFORMANCE_INDICATOR


def get_performance_indicator(name, *args, d={}, **kwargs):
    return get_from_list(get_performance_indicator_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# DECOMPOSITION
# =========================================================================================================

def get_decomposition_options():
    from pymoo.decomposition.pbi import PBI
    from pymoo.decomposition.tchebicheff import Tchebicheff
    from pymoo.decomposition.weighted_sum import WeightedSum
    from pymoo.decomposition.asf import ASF
    from pymoo.decomposition.aasf import AASF
    from pymoo.decomposition.perp_dist import PerpendicularDistance

    DECOMPOSITION = [
        ("weighted-sum", WeightedSum),
        ("tchebi", Tchebicheff),
        ("pbi", PBI),
        ("asf", ASF),
        ("aasf", AASF),
        ("perp_dist", PerpendicularDistance)
    ]

    return DECOMPOSITION


def get_decomposition(name, *args, d={}, **kwargs):
    return get_from_list(get_decomposition_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# DECOMPOSITION
# =========================================================================================================

def get_decision_making_options():
    from pymoo.decision_making.high_tradeoff import HighTradeoffPoints
    from pymoo.decision_making.pseudo_weights import PseudoWeights

    DECISION_MAKING = [
        ("high-tradeoff", HighTradeoffPoints),
        ("pseudo-weights", PseudoWeights)
    ]

    return DECISION_MAKING


def get_decision_making(name, *args, d={}, **kwargs):
    return get_from_list(get_decision_making_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Documentation
# =========================================================================================================


def dummy(name, kwargs):
    """
    A convenience method to get a {type} object just by providing a string.

    Parameters
    ----------

    name : {{ {options} }}
        Name of the {type}.

    kwargs : dict
        Dictionary that should be used to call the method mapped to the {type} factory function.

    Returns
    -------
    class : {clazz}
        An {type} object based on the string. `None` if the {type} was not found.

    """
    pass


def options_to_string(l):
    return ", ".join(["'%s'" % k[0] for k in l])


if Configuration.parse_custom_docs:
    from pymoo.docs import parse_doc_string

    from pymoo.factory import get_algorithm_options, get_selection_options, get_crossover_options, \
        get_mutation_options, get_termination_options, get_algorithm, get_selection, get_crossover, get_mutation, \
        get_termination, get_sampling, get_sampling_options

    parse_doc_string(dummy, get_algorithm, {"type": "algorithm",
                                            "clazz": ":class:`~pymoo.model.algorithm.Algorithm`",
                                            "options": options_to_string(get_algorithm_options())
                                            })

    parse_doc_string(dummy, get_sampling, {"type": "sampling",
                                           "clazz": ":class:`~pymoo.model.sampling.Sampling`",
                                           "options": options_to_string(get_sampling_options())
                                           })

    parse_doc_string(dummy, get_selection, {"type": "selection",
                                            "clazz": ":class:`~pymoo.model.selection.Selection`",
                                            "options": options_to_string(get_selection_options())
                                            })

    parse_doc_string(dummy, get_crossover, {"type": "crossover",
                                            "clazz": ":class:`~pymoo.model.crossover.Crossover`",
                                            "options": options_to_string(get_crossover_options())
                                            })

    parse_doc_string(dummy, get_mutation, {"type": "mutation",
                                           "clazz": ":class:`~pymoo.model.mutation.Mutation`",
                                           "options": options_to_string(get_mutation_options())
                                           })

    parse_doc_string(dummy, get_termination, {"type": "termination",
                                              "clazz": ":class:`~pymoo.model.termination.termination`",
                                              "options": options_to_string(get_termination_options())
                                              })



from pymoo.model.sampling import Sampling

import numpy as np

from pymoo.util.normalization import denormalize


class FloatRandomSampling(Sampling):
    """
    Randomly sample points in the real space by considering the lower and upper bounds of the problem.
    """

    def __init__(self, var_type=np.float) -> None:
        super().__init__()
        self.var_type = var_type

    def _do(self, problem, n_samples, **kwargs):
        val = np.random.random((n_samples, problem.n_var))
        return denormalize(val, problem.xl, problem.xu)


class BinaryRandomSampling(Sampling):

    def __init__(self) -> None:
        super().__init__()

    def _do(self, problem, n_samples, **kwargs):
        val = np.random.random((n_samples, problem.n_var))
        return (val < 0.5).astype(np.bool)



from __future__ import division, print_function, absolute_import

import numpy as np
from numpy import (zeros, array, asfarray, concatenate)
from scipy.optimize._slsqp import slsqp

from pymoo.algorithms.so_local_search import LocalSearch
from pymoo.model.individual import Individual
from pymoo.model.population import Population
from pymoo.model.problem import Problem
from pymoo.util.approx_grad import GradientApproximation
from pymoo.util.display import SingleObjectiveDisplay


class SQLPDisplay(SingleObjectiveDisplay):

    def _do(self, problem, evaluator, algorithm):
        if algorithm.major:
            super()._do(problem, evaluator, algorithm)
            # self.output.append("iter", algorithm.D["majiter"])


class SQLP(LocalSearch):

    def __init__(self,
                 n_evals=1000,
                 ftol=1e-6,
                 display=SQLPDisplay(),
                 **kwargs):

        super().__init__(display=display, **kwargs)
        self.ftol = ftol
        self.n_evals = n_evals
        self.D = None
        self.pop = None
        self.majiter_prev = 1

        self.args = ["m", "meq", "X", "xl", "xu", "F", "G", "dF", "dG", "acc", "majiter", "mode", "w", "jw",
                     "alpha", "f0", "gs", "h1", "h2", "h3", "h4", "t", "t0", "tol",
                     "iexact", "incons", "ireset", "itermx", "line", "n1", "n2", "n3"]

    def initialize(self, problem, **kwargs):
        super().initialize(problem, **kwargs)

        n, meq, mieq = problem.n_var, 0, problem.n_constr
        m = meq + mieq
        la = max(1, m)

        n1 = n + 1
        mineq = m - meq + n1 + n1
        len_w = (3 * n1 + m) * (n1 + 1) + (n1 - meq + 1) * (mineq + 2) + 2 * mineq + (n1 + mineq) * (n1 - meq) \
                + 2 * meq + n1 + ((n + 1) * n) // 2 + 2 * m + 3 * n + 3 * n1 + 1
        len_jw = mineq

        self.D = dict(
            X=None,
            n=n,
            meq=meq,
            mieq=mieq,
            la=la,
            m=m,
            xl=problem.xl.astype(np.float),
            xu=problem.xu.astype(np.float),
            mode=array(0, int),
            acc=array(self.ftol, float),
            majiter=array(self.n_evals, int),
            majiter_prev=0,
            alpha=array(0, float),
            f0=array(0, float),
            gs=array(0, float),
            h1=array(0, float),
            h2=array(0, float),
            h3=array(0, float),
            h4=array(0, float),
            t=array(0, float),
            t0=array(0, float),
            tol=array(0, float),
            iexact=array(0, int),
            incons=array(0, int),
            ireset=array(0, int),
            itermx=array(0, int),
            line=array(0, int),
            n1=array(0, int),
            n2=array(0, int),
            n3=array(0, int),
            w=zeros(len_w),
            jw=zeros(len_jw)
        )

    def _initialize(self):
        super()._initialize()

        # Clip initial guess to bounds (SLSQP may fail with bounds-infeasible initial point)
        x = asfarray(self.x0.X.flatten())
        xl, xu = self.problem.bounds()
        have_bound = np.isfinite(xl)
        x[have_bound] = np.clip(x[have_bound], xl[have_bound], np.inf)
        have_bound = np.isfinite(xu)
        x[have_bound] = np.clip(x[have_bound], -np.inf, xu[have_bound])
        self.D["X"] = x

        self.pop = Population()
        self._eval_obj()
        self._eval_grad()
        self._update()
        self._call()

        self.major = True

    def _update(self):
        D = self.D
        ind = Individual(X=np.copy(D["X"]), F=np.copy(D["F"]), G=np.copy(-D["G"]))
        ind.CV = Problem.calc_constraint_violation(ind.G[None, :])[0]
        ind.feasible = (ind.CV <= 0)
        self.pop = Population.merge(self.pop, Population.create(ind))

    def _call(self):
        # args = np.load("/Users/blankjul/workspace/pymoo/pymoo/algorithms/data_constr.npy", allow_pickle=True)
        #
        # pymoo = [self.D[e] for e in self.args]
        # for k in range(len(self.args)):
        #     orig = tuple(args)[k]
        #     _pymoo = pymoo[k]
        #     if isinstance(orig, np.ndarray):
        #         print(orig.dtype, _pymoo.dtype)
        #         print(orig.shape, _pymoo.shape)
        #     print(type(orig), type(_pymoo))
        #     print(orig)
        #     print(_pymoo)
        #
        #     print()
        #
        # print("-------------------------------------------------")

        slsqp(*[self.D[e] for e in self.args])

    def _step(self):
        self._eval_obj()
        self._update()

        self._call()

        if self.D["mode"] == -1:
            self._eval_grad()
            self._call()

    def _eval_obj(self):
        D = self.D
        D["F"], _, G = self.evaluator.eval(self.problem, D["X"])
        D["G"] = - G if self.problem.n_constr > 0 else np.zeros(0)

    def _eval_grad(self):
        D = self.D

        # dF, dG = self.problem.evaluate(D["X"], return_values_of=["dF", "dG"])
        # if dF is None or dG is None:
        #     dF, dG = GradientApproximation(self.problem, evaluator=self.evaluator).do(Individual(X=D["X"]))

        dF, dG = GradientApproximation(self.problem, evaluator=self.evaluator).do(Individual(X=D["X"]))

        D["dF"] = concatenate([dF[0], [0]])
        if self.problem.n_constr > 0:
            D["dG"] = - np.column_stack([dG, zeros(self.problem.n_constr)])
        else:
            D["dG"] = zeros((1, self.problem.n_var + 1))

    def _next(self):
        self._step()

        majiter = int(self.D["majiter"])
        if abs(self.D["mode"]) != 1:
            self.termination.force_termination = True

        self.major = self.majiter_prev < majiter
        self.majiter_prev = majiter



import numpy as np

from pymoo.algorithms.so_gradient_descent import GradientBasedAlgorithm


class Adam(GradientBasedAlgorithm):

    def __init__(self,
                 X,
                 alpha=0.005,
                 beta_1=0.9,
                 beta_2=0.999,
                 epsilon=1e-8,
                 **kwargs) -> None:
        super().__init__(X, **kwargs)

        self.alpha = alpha
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon

        self.t = 0
        self.m_t = 0
        self.v_t = 0

    def apply(self):
        X, dX = self.X, self.dX

        self.t += 1
        beta_1, beta_2 = self.beta_1, self.beta_2

        # update moving average of gradient and squared gradient
        self.m_t = beta_1 * self.m_t + (1 - beta_1) * dX
        self.v_t = beta_2 * self.v_t + (1 - beta_2) * (dX * dX)

        # calculates the bias-corrected estimates
        m_cap = self.m_t / (1 - (beta_1 ** self.t))
        v_cap = self.v_t / (1 - (beta_2 ** self.t))

        # do the gradient update
        self.X = X - (self.alpha * m_cap) / (np.sqrt(v_cap) + self.epsilon)

    def restart(self):
        self.t = 0
        self.m_t = 0
        self.v_t = 0
        self.alpha /= 2





from pymoo.algorithms.so_de import DE
from pymoo.factory import get_problem
from pymoo.operators.sampling.latin_hypercube_sampling import LatinHypercubeSampling
from pymoo.optimize import minimize


problem = get_problem("ackley", n_var=10)

algorithm = DE(
    pop_size=100,
    sampling=LatinHypercubeSampling(iterations=100, criterion="maxmin"),
    variant="DE/rand/1/bin",
    CR=0.5,
    F=0.3,
    dither="vector",
    jitter=False
)

res = minimize(problem,
               algorithm,
               seed=1,
               verbose=False)

print("Best solution found: \nX = %s\nF = %s" % (res.X, res.F))


import multiprocessing
import warnings
from abc import abstractmethod
from multiprocessing.pool import ThreadPool

import autograd
import autograd.numpy as anp
import numpy as np

from pymoo.problems.gradient import run_and_trace, calc_jacobian
from pymoo.util.misc import at_least_2d_array


class Problem:
    """
    Superclass for each problem that is defined. It provides attributes such
    as the number of variables, number of objectives or constraints.
    Also, the lower and upper bounds are stored. If available the Pareto-front, nadir point
    and ideal point are stored.

    """

    def __init__(self,
                 n_var=-1,
                 n_obj=-1,
                 n_constr=0,
                 xl=None,
                 xu=None,
                 type_var=np.double,
                 evaluation_of="auto",
                 parallelization=None,
                 elementwise_evaluation=False,
                 callback=None):
        """

        Parameters
        ----------
        n_var : int
            number of variables
        n_obj : int
            number of objectives
        n_constr : int
            number of constraints
        xl : np.array or int
            lower bounds for the variables. if integer all lower bounds are equal.
        xu : np.array or int
            upper bounds for the variable. if integer all upper bounds are equal.
        type_var : numpy type
            type of the variable to be evaluated. Can also be np.object if it is a complex data type
        elementwise_evaluation : bool

        parallelization : str or tuple
            See :ref:`nb_parallelization` for guidance on parallelization.

        """

        # number of variable for this problem
        self.n_var = n_var

        # type of the variable to be evaluated
        self.type_var = type_var

        # number of objectives
        self.n_obj = n_obj

        # number of constraints
        self.n_constr = n_constr

        # allow just an integer for xl and xu if all bounds are equal
        if n_var > 0 and not isinstance(xl, np.ndarray) and xl is not None:
            self.xl = np.ones(n_var) * xl
        else:
            self.xl = xl

        if n_var > 0 and not isinstance(xu, np.ndarray) and xu is not None:
            self.xu = np.ones(n_var) * xu
        else:
            self.xu = xu

        # the pareto set and front will be calculated only once and is stored here
        self._pareto_front = None
        self._pareto_set = None
        self._ideal_point, self._nadir_point = None, None

        # actually defines what _evaluate is setting during the evaluation
        if evaluation_of == "auto":
            # by default F is set, and G if the problem does have constraints
            self.evaluation_of = ["F"]
            if self.n_constr > 0:
                self.evaluation_of.append("G")
        else:
            self.evaluation_of = evaluation_of

        # whether the evaluation function is called per set of solutions or single solution
        self.elementwise_evaluation = elementwise_evaluation

        # only applicable if elementwise_evaluation is true - if, how should the single evaluations be parallelized
        self.parallelization = parallelization

        # store the callback if defined
        self.callback = callback

    def nadir_point(self):
        """
        Returns
        -------
        nadir_point : np.array
            The nadir point for a multi-objective problem.
            If single-objective, it returns the best possible solution which is equal to the ideal point.

        """
        # if the ideal point has not been calculated yet
        if self._nadir_point is None:

            # calculate the pareto front if not happened yet
            if self._pareto_front is None:
                self.pareto_front()

            # if already done or it was successful - calculate the ideal point
            if self._pareto_front is not None:
                self._ideal_point = np.max(self._pareto_front, axis=0)

        return self._nadir_point

    def ideal_point(self):
        """
        Returns
        -------
        ideal_point : np.array
            The ideal point for a multi-objective problem. If single-objective
            it returns the best possible solution.
        """

        # if the ideal point has not been calculated yet
        if self._ideal_point is None:

            # calculate the pareto front if not happened yet
            if self._pareto_front is None:
                self.pareto_front()

            # if already done or it was successful - calculate the ideal point
            if self._pareto_front is not None:
                self._ideal_point = np.min(self._pareto_front, axis=0)

        return self._ideal_point

    def pareto_front(self, *args, use_cache=True, exception_if_failing=True, **kwargs):
        """
        Parameters
        ----------

        args : Same problem implementation need some more information to create the Pareto front. For instance
                the DTLZ problem suite generates the Pareto front by usage of the reference directions.
                We refer to the corresponding problem for more information.
        exception_if_failing : bool
                Whether to throw an exception when generating the Pareto front has failed.
        use_cache : bool
                Whether to use the cache if the Pareto front has been generated beforehand.

        Returns
        -------
        P : np.array
            The Pareto front of a given problem. It is only loaded or calculate the first time and then cached.
            For a single-objective problem only one point is returned but still in a two dimensional array.

        """
        if not use_cache or self._pareto_front is None:
            try:
                pf = self._calc_pareto_front(*args, **kwargs)
                if pf is not None:
                    self._pareto_front = at_least_2d_array(pf)

            except Exception as e:
                if exception_if_failing:
                    raise e

        return self._pareto_front

    def pareto_set(self, *args, use_cache=True, **kwargs):
        """
        Returns
        -------
        S : np.array
            Returns the pareto set for a problem. Points in the X space to be known to be optimal!
        """
        if not use_cache or self._pareto_set is None:
            self._pareto_set = at_least_2d_array(self._calc_pareto_set(*args, **kwargs))

        return self._pareto_set

    def evaluate(self,
                 X,
                 *args,
                 return_values_of="auto",
                 return_as_dictionary=False,
                 **kwargs):

        """
        Evaluate the given problem.

        The function values set as defined in the function.
        The constraint values are meant to be positive if infeasible. A higher positive values means "more" infeasible".
        If they are 0 or negative, they will be considered as feasible what ever their value is.

        Parameters
        ----------

        X : np.array
            A two dimensional matrix where each row is a point to evaluate and each column a variable.

        return_as_dictionary : bool
            If this is true than only one object, a dictionary, is returned. This contains all the results
            that are defined by return_values_of. Otherwise, by default a tuple as defined is returned.

        return_values_of : list of strings
            You can provide a list of strings which defines the values that are returned. By default it is set to
            "auto" which means depending on the problem the function values or additional the constraint violation (if
            the problem has constraints) are returned. Otherwise, you can provide a list of values to be returned.

            Allowed is ["F", "CV", "G", "dF", "dG", "dCV", "feasible"] where the d stands for
            derivative and h stands for hessian matrix.


        Returns
        -------

            A dictionary, if return_as_dictionary enabled, or a list of values as defined in return_values_of.

        """

        # call the callback of the problem
        if self.callback is not None:
            self.callback(X)

        # make the array at least 2-d - even if only one row should be evaluated
        only_single_value = len(np.shape(X)) == 1
        X = np.atleast_2d(X)

        # check the dimensionality of the problem and the given input
        if X.shape[1] != self.n_var:
            raise Exception('Input dimension %s are not equal to n_var %s!' % (X.shape[1], self.n_var))

        # automatic return the function values and CV if it has constraints if not defined otherwise
        if type(return_values_of) == str and return_values_of == "auto":
            return_values_of = ["F"]
            if self.n_constr > 0:
                return_values_of.append("CV")

        # all values that are set in the evaluation function
        values_not_set = [val for val in return_values_of if val not in self.evaluation_of]

        # have a look if gradients are not set and try to use autograd and calculate grading if implemented using it
        gradients_not_set = [val for val in values_not_set if val.startswith("d")]

        # whether gradient calculation is necessary or not
        calc_gradient = (len(gradients_not_set) > 0)

        # set in the dictionary if the output should be calculated - can be used for the gradient
        out = {}
        for val in return_values_of:
            out[val] = None

        # calculate the output array - either elementwise or not. also consider the gradient
        if self.elementwise_evaluation:
            out = self._evaluate_elementwise(X, calc_gradient, out, *args, **kwargs)
        else:
            out = self._evaluate_batch(X, calc_gradient, out, *args, **kwargs)

            calc_gradient_of = [key for key, val in out.items()
                                if "d" + key in return_values_of and
                                out.get("d" + key) is None and
                                (type(val) == autograd.numpy.numpy_boxes.ArrayBox)]

            if len(calc_gradient_of) > 0:
                deriv = self._calc_gradient(out, calc_gradient_of)
                out = {**out, **deriv}

        # convert back to conventional numpy arrays - no array box as return type
        for key in out.keys():
            if type(out[key]) == autograd.numpy.numpy_boxes.ArrayBox:
                out[key] = out[key]._value

        # if constraint violation should be returned as well
        if self.n_constr == 0:
            CV = np.zeros([X.shape[0], 1])
        else:
            CV = Problem.calc_constraint_violation(out["G"])

        if "CV" in return_values_of:
            out["CV"] = CV

        # if an additional boolean flag for feasibility should be returned
        if "feasible" in return_values_of:
            out["feasible"] = (CV <= 0)

        # if asked for a value but not set in the evaluation set to None
        for val in return_values_of:
            if val not in out:
                out[val] = None

        # remove the first dimension of the output - in case input was a 1d- vector
        if only_single_value:
            for key in out.keys():
                if out[key] is not None:
                    out[key] = out[key][0, :]

        if return_as_dictionary:
            return out
        else:

            # if just a single value do not return a tuple
            if len(return_values_of) == 1:
                return out[return_values_of[0]]
            else:
                return tuple([out[val] for val in return_values_of])

    def _calc_gradient(self, out, keys):

        deriv = {}
        for key in keys:
            val = out[key]

            # calculate the jacobian matrix and set it - (ignore warnings of autograd here)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                jac = calc_jacobian(out["__autograd__"], val)
                deriv["d" + key] = jac

        return deriv

    def _evaluate_batch(self, X, calc_gradient, out, *args, **kwargs):
        if calc_gradient:
            out["__autograd__"], _ = run_and_trace(self._evaluate, X, *[out])
        else:
            self._evaluate(X, out, *args, **kwargs)
        at_least2d(out)

        return out

    def _evaluate_elementwise(self, X, calc_gradient, out, *args, **kwargs):
        ret = []

        def func(_x):
            _out = {}
            if calc_gradient:
                grad, _ = run_and_trace(self._evaluate, _x, *[_out])
                _out["__autograd__"] = grad
            else:
                self._evaluate(_x, _out, *args, **kwargs)
            return _out

        parallelization = self.parallelization
        if not isinstance(parallelization, (list, tuple)):
            parallelization = [self.parallelization]

        _type = parallelization[0]
        if len(parallelization) >= 1:
            _params = parallelization[1:]

        # just serialize evaluation
        if _type is None:
            [ret.append(func(x)) for x in X]

        elif _type is "starmap":
            if len(_params) != 1:
                raise Exception("The starmap parallelization method must be accompanied by a starmapping callable")

            params = [[X[k], calc_gradient, self._evaluate, args, kwargs] for k in range(len(X))]

            starmapper = _params[0]
            ret = list(starmapper(evaluate_in_parallel, params))

        elif _type == "threads":

            if len(_params) == 0:
                n_threads = multiprocessing.cpu_count() - 1
            else:
                n_threads = _params[0]

            with ThreadPool(n_threads) as pool:
                params = [[X[k], calc_gradient, self._evaluate, args, kwargs] for k in range(len(X))]
                ret = pool.starmap(evaluate_in_parallel, params)

        elif _type == "dask":

            if len(_params) != 2:
                raise Exception("A distributed client objective is need for using dask. parallelization=(dask, "
                                "<client>, <function>).")
            else:
                client, fun = _params

            jobs = []
            for k in range(len(X)):
                jobs.append(client.submit(fun, X[k]))

            ret = [job.result() for job in jobs]

        else:
            raise Exception("Unknown parallelization method: %s (should be one of: None, starmap, threads, dask)" % _type)

        # stack all the single outputs together
        for key in ret[0].keys():
            out[key] = anp.row_stack([ret[i][key] for i in range(len(ret))])

        return out

    @abstractmethod
    def _evaluate(self, x, out, *args, **kwargs):
        pass

    def has_bounds(self):
        return self.xl is not None and self.xu is not None

    def bounds(self):
        return self.xl, self.xu

    def name(self):
        """
        Returns
        -------
        name : str
            The name of the problem. Per default it is the name of the class but it can be overridden.
        """
        return self.__class__.__name__

    def _calc_pareto_front(self, *args, **kwargs):
        """
        Method that either loads or calculates the pareto front. This is only done
        ones and the pareto front is stored.

        Returns
        -------
        pf : np.array
            Pareto front as array.

        """
        pass

    def _calc_pareto_set(self, *args, **kwargs):
        pass

    # some problem information
    def __str__(self):
        s = "# name: %s\n" % self.name()
        s += "# n_var: %s\n" % self.n_var
        s += "# n_obj: %s\n" % self.n_obj
        s += "# n_constr: %s\n" % self.n_constr
        s += "# f(xl): %s\n" % self.evaluate(self.xl)[0]
        s += "# f((xl+xu)/2): %s\n" % self.evaluate((self.xl + self.xu) / 2.0)[0]
        s += "# f(xu): %s\n" % self.evaluate(self.xu)[0]
        return s

    @staticmethod
    def calc_constraint_violation(G):
        if G is None:
            return None
        elif G.shape[1] == 0:
            return np.zeros(G.shape[0])[:, None]
        else:
            return np.sum(G * (G > 0).astype(np.float), axis=1)[:, None]

    def __getstate__(self):
        state = self.__dict__.copy()

        # If the parallelization is starmap,
        # we can't pickle self with the starmapper function.
        if state["parallelization"] is not None and state["parallelization"][0] == "starmap":
            state["parallelization"] = None

        return state

# makes all the output at least 2-d dimensional
def at_least2d(d):
    for key in d.keys():
        if len(np.shape(d[key])) == 1:
            d[key] = d[key][:, None]


def evaluate_in_parallel(_x, calc_gradient, func, args, kwargs):
    _out = {}
    if calc_gradient:
        _out["__autograd__"], _ = run_and_trace(func, _x, *[_out])
    else:
        func(_x, _out, *args, **kwargs)
    return _out


def evaluate_in_parallel_object(_x, calc_gradient, obj, args, kwargs):
    _out = {}
    obj._evaluate(_x, _out, *args, **kwargs)
    return _out


def get_problem_from_func(func, xl=None, xu=None, n_var=None, func_args={}):
    if xl is None or xu is None:
        raise Exception("Please provide lower and upper bounds for the problem.")
    if isinstance(xl, (int, float)):
        xl = xl * anp.ones(n_var)
    if isinstance(xu, (int, float)):
        xu = xu * anp.ones(n_var)

    # determine through a test evaluation details about the problem
    n_var = xl.shape[0]
    n_obj = -1
    n_constr = 0

    out = {}
    func(xl[None, :], out, **func_args)
    at_least2d(out)

    n_obj = out["F"].shape[1]
    if out.get("G") is not None:
        n_constr = out["G"].shape[1]

    class MyProblem(Problem):
        def __init__(self):
            Problem.__init__(self)
            self.n_var = n_var
            self.n_constr = n_constr
            self.n_obj = n_obj
            self.func = self._evaluate
            self.xl = xl
            self.xu = xu

        def _evaluate(self, x, out, *args, **kwargs):
            func(x, out, *args, **kwargs)

    return MyProblem()



from pymoo.model.termination import Termination
from pymoo.performance_indicator.igd import IGD


class IGDTermination(Termination):

    def __init__(self, min_igd, pf) -> None:
        super().__init__()
        if pf is None:
            raise Exception("You can only use IGD termination criteria if the pareto front is known!")

        self.perf = IGD(pf)
        self.min_igd = min_igd

    def _do_continue(self, algorithm):
        F = algorithm.opt.get("F")
        return self.perf.calc(F) > self.min_igd




import autograd.numpy as anp

from pymoo.model.problem import Problem


def curve(problem, n_points=200):
    X = anp.linspace(problem.xl[0], problem.xu[0], n_points)[:, None]
    F = problem.evaluate(X)
    return anp.column_stack([X, F])


class MultiModalSimple1(Problem):
    def __init__(self, n_var=1):
        super().__init__(n_var=n_var, n_obj=1, n_constr=0, xl=0, xu=1, type_var=anp.double)

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = 1 - anp.exp(-x ** 2) * anp.sin(2 * anp.pi * x) ** 2


class MultiModalSimple2(Problem):
    def __init__(self, n_var=1):
        super().__init__(n_var=n_var, n_obj=1, n_constr=0, xl=-1, xu=0, type_var=anp.double)

    def _evaluate(self, x, out, *args, **kwargs):
        x = - x
        out["F"] = 1.1 - anp.exp(-2 * x) * anp.sin(5 * anp.pi * x) ** 2



# START load_data
import numpy as np
from pymoo.factory import get_problem
from pymoo.visualization.scatter import Scatter

# The pareto front of a scaled zdt1 problem
pf = get_problem("zdt1").pareto_front()

# The result found by an algorithm
A = pf[::10] * 1.1

# plot the result
Scatter(legend=True).add(pf, label="Pareto-front").add(A, label="Result").show()
# END load_data





# START gd
from pymoo.factory import get_performance_indicator

gd = get_performance_indicator("gd", pf)
print("GD", gd.calc(A))
# END gd



# START gd_plus
from pymoo.factory import get_performance_indicator

gd_plus = get_performance_indicator("gd+", pf)
print("GD+", gd_plus.calc(A))
# END gd_plus


# START igd
from pymoo.factory import get_performance_indicator

igd = get_performance_indicator("igd", pf)
print("IGD", igd.calc(A))
# END igd



# START igd_plus
from pymoo.factory import get_performance_indicator

igd_plus = get_performance_indicator("igd+", pf)
print("IGD+", igd_plus.calc(A))
# END igd_plus


# START hv
from pymoo.factory import get_performance_indicator

hv = get_performance_indicator("hv", ref_point=np.array([1.2, 1.2]))
print("hv", hv.calc(A))
# END hv


import numpy as np

from pymoo.model.individual import Individual


class Population(np.ndarray):

    def __new__(cls, n_individuals=0, individual=Individual()):
        obj = super(Population, cls).__new__(cls, n_individuals, dtype=individual.__class__).view(cls)
        for i in range(n_individuals):
            obj[i] = individual.copy()
        obj.individual = individual
        return obj

    @classmethod
    def merge(cls, a, b):
        a, b = pop_from_array_or_individual(a), pop_from_array_or_individual(b)
        return a.merge(b)

    def merge(self, other):
        other = pop_from_array_or_individual(other)
        if len(self) == 0:
            return other
        else:
            obj = np.concatenate([self, other]).view(Population)
            obj.individual = self.individual
            return obj

    def copy(self):
        pop = Population(n_individuals=len(self), individual=self.individual)
        for i in range(len(self)):
            pop[i] = self[i]
        return pop

    def has(self, key):
        return all([ind.has(key) for ind in self])

    def __deepcopy__(self, memo):
        return self.copy()

    @classmethod
    def create(cls, *args):
        pop = np.concatenate([pop_from_array_or_individual(arg) for arg in args]).view(Population)
        pop.individual = Individual()
        return pop

    def new(self, *args):

        if len(args) == 1:
            return Population(n_individuals=args[0], individual=self.individual)
        else:
            n = len(args[1]) if len(args) > 0 else 0
            pop = Population(n_individuals=n, individual=self.individual)
            if len(args) > 0:
                pop.set(*args)
            return pop

    def collect(self, func, to_numpy=True):
        val = []
        for i in range(len(self)):
            val.append(func(self[i]))
        if to_numpy:
            val = np.array(val)
        return val

    def set(self, *args):

        for i in range(int(len(args) / 2)):

            key, values = args[i * 2], args[i * 2 + 1]
            is_iterable = hasattr(values, '__len__') and not isinstance(values, str)

            if is_iterable and len(values) != len(self):
                raise Exception("Population Set Attribute Error: Number of values and population size do not match!")

            for i in range(len(self)):
                val = values[i] if is_iterable else values
                self[i].set(key, val)

        return self

    def get(self, *args, to_numpy=True):

        val = {}
        for c in args:
            val[c] = []

        for i in range(len(self)):

            for c in args:
                val[c].append(self[i].get(c))

        res = [val[c] for c in args]
        if to_numpy:
            res = [np.array(e) for e in res]

        if len(args) == 1:
            return res[0]
        else:
            return tuple(res)

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.individual = getattr(obj, 'individual', None)


def pop_from_array_or_individual(array, pop=None):
    # the population type can be different - (different type of individuals)
    if pop is None:
        pop = Population()

    # provide a whole population object - (individuals might be already evaluated)
    if isinstance(array, Population):
        pop = array
    elif isinstance(array, np.ndarray):
        pop = pop.new("X", np.atleast_2d(array))
    elif isinstance(array, Individual):
        pop = Population(1)
        pop[0] = array
    else:
        return None

    return pop

if __name__ == '__main__':
    pop = Population(10)
    pop.get("F")
    pop.new()
    print("")



# START rnsga3
import numpy as np

from pymoo.algorithms.rnsga3 import RNSGA3
from pymoo.factory import get_problem
from pymoo.optimize import minimize
from pymoo.visualization.scatter import Scatter

problem = get_problem("zdt1")
pf = problem.pareto_front()

# Define reference points
ref_points = np.array([[0.3, 0.4], [0.8, 0.5]])

# Get Algorithm
algorithm = RNSGA3(
    ref_points=ref_points,
    pop_per_ref_point=50,
    mu=0.1)

res = minimize(problem,
               algorithm=algorithm,
               termination=('n_gen', 300),
               pf=pf,
               seed=1,
               verbose=False)

reference_directions = res.algorithm.survival.ref_dirs

plot = Scatter()
plot.add(pf, label="pf")
plot.add(res.F, label="F")
plot.add(ref_points, label="ref_points")
plot.add(reference_directions, label="ref_dirs")
plot.show()

# END rnsga3

# START rnsga3_3d
from pymoo.util.reference_direction import UniformReferenceDirectionFactory

# Get problem
problem = get_problem("dtlz4", n_var=12, n_obj=3)

# Define reference points and reference directions
ref_points = np.array([[1.0, 0.5, 0.2], [0.3, 0.2, 0.6]])
ref_dirs = UniformReferenceDirectionFactory(3, n_points=91).do()
pf = problem.pareto_front(ref_dirs)

# Get Algorithm
algorithm = RNSGA3(
    ref_points=ref_points,
    pop_per_ref_point=91,
    mu=0.1)

res = minimize(problem,
               algorithm,
               termination=('n_gen', 300),
               pf=pf,
               seed=1,
               verbose=False)


plot = Scatter()
plot.add(pf, label="pf")
plot.add(res.F, label="F")
plot.add(ref_points, label="ref_points")
plot.show()

# END rnsga3_3d


import numpy as np

from pymoo.algorithms.so_local_search import LocalSearch
from pymoo.model.individual import Individual
from pymoo.model.population import Population
from pymoo.model.problem import Problem
from pymoo.optimize import minimize
from pymoo.problems.single import Himmelblau, Sphere, Rastrigin
from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.util.normalization import normalize, denormalize


def norm_bounds(pop, problem):
    nxl = normalize(pop.get("xl"), problem.xl, problem.xu)
    nxu = normalize(pop.get("xu"), problem.xl, problem.xu)
    return nxl, nxu


def update_bounds(ind, xl, xu, k, delta):
    _xl = np.copy(xl)
    _xl[k] = ind.X[k] - delta
    ind.set("xl", _xl)

    _xu = np.copy(xu)
    _xu[k] = ind.X[k] + delta
    ind.set("xu", _xu)


class DIRECT(LocalSearch):

    def __init__(self,
                 eps=1e-2,
                 penalty=0.1,
                 display=SingleObjectiveDisplay(),
                 **kwargs):
        super().__init__(display=display, **kwargs)
        self.eps = eps
        self.penalty = penalty

    def initialize(self, problem, **kwargs):
        super().initialize(problem, **kwargs)

        xl, xu = problem.bounds()
        X = denormalize(0.5 * np.ones(problem.n_var), xl, xu)
        x0 = Individual(X=X)
        x0.set("xl", xl)
        x0.set("xu", xu)
        x0.set("depth", 0)
        self.x0 = x0

    def _initialize(self, **kwargs):
        super()._initialize(**kwargs)

    def _potential_optimal(self):
        pop = self.pop

        if len(pop) == 1:
            return pop

        # get the intervals of each individual
        _F, _CV, xl, xu = pop.get("F", "CV", "xl", "xu")
        nF = normalize(_F)
        F = nF + self.penalty * _CV

        # get the length of the interval of each solution
        nxl, nxu = norm_bounds(pop, problem)
        length = (nxu - nxl) / 2

        val = length.max(axis=1)

        # (a) non-dominated with respect to interval
        obj = np.column_stack([-val, F])
        I = NonDominatedSorting().do(obj, only_non_dominated_front=True)
        candidates, F, xl, xu, val = pop[I], F[I], xl[I], xu[I], val[I]

        # import matplotlib.pyplot as plt
        # plt.scatter(obj[:, 0], obj[:, 1])
        # plt.scatter(obj[I, 0], obj[I, 1], color="red")
        # plt.show()

        if len(candidates) == 1:
            return candidates

        else:
            # TODO: The second condition needs to be implemented here. Exact implementation still unclear.

            n_max_candidates = 10

            if len(candidates) > n_max_candidates:
                I = list(np.random.choice(np.arange(len(candidates)), n_max_candidates - 1))
                k = np.argmin(F[:, 0])
                if k not in I:
                    I.append(k)
                candidates = candidates[I]

            return candidates

    def _next(self):
        # the offspring population to finally evaluate and attach to the population
        off = Population()

        # find the potential optimal solution in the current population
        potential_optimal = self._potential_optimal()

        # for each of those solutions execute the division move
        for current in potential_optimal:

            # find the largest dimension the solution has not been evaluated yet
            nxl, nxu = norm_bounds(current, problem)
            k = np.argmax(nxu - nxl)

            # the delta value to be used to get left and right - this is one sixth of the range
            xl, xu = current.get("xl"), current.get("xu")

            delta = (xu[k] - xl[k]) / 6

            # print(current.X, delta, k, xl, xu)

            # create the left individual
            left_x = np.copy(current.X)
            left_x[k] = xl[k] + delta
            left = Individual(X=left_x)

            # create the right individual
            right_x = np.copy(current.X)
            right_x[k] = xu[k] - delta
            right = Individual(X=right_x)

            # update the boundaries for all the points accordingly
            for ind in [current, left, right]:
                update_bounds(ind, xl, xu, k, delta)

            # create the offspring population, evaluate and attach to current population
            _off = Population.create(left, right)
            _off.set("depth", current.get("depth") + 1)

            off = Population.merge(off, _off)

        # evaluate the offsprings
        self.evaluator.eval(self.problem, off, algorithm=self)

        # print(off.get("X"))

        # add the offsprings to the population
        self.pop = Population.merge(self.pop, off)


class ExSwarm(Problem):

    def __init__(self):
        super().__init__(n_var=2, n_obj=1, n_constr=0, xl=np.array([-1.0, -1.0]), xu=np.array([1.0, 1.0]))

    def _evaluate(self, x, out, *args, **kwargs):
        v1 = 20 * x[:, 0]
        v2 = 20 * x[:, 1]
        out["F"] = -np.sin(v1 / np.pi) ** 2. * np.sin(v2 / np.pi) ** 2. * (abs(v1) + abs(v2) + 0.1 * (v1 + v2)) + (
                v1 * v1 + v2 * v2) / 30


if __name__ == '__main__':
    problem = ExSwarm()
    problem = Himmelblau()
    # problem = Sphere(n_var=100, opt=0.2 * np.ones(100))
    problem = Rastrigin(n_var=10)
    problem.xl *= 1.5

    # problem = get_problem("g02")
    algorithm = DIRECT()
    # algorithm = GA()

    ret = minimize(problem,
                   algorithm,
                   ("n_iter", 1000),
                   verbose=True)



import numpy as np

from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm
from pymoo.docs import parse_doc_string
from pymoo.model.individual import Individual
from pymoo.model.survival import Survival
from pymoo.operators.crossover.simulated_binary_crossover import SimulatedBinaryCrossover
from pymoo.operators.mutation.polynomial_mutation import PolynomialMutation
from pymoo.operators.sampling.random_sampling import FloatRandomSampling
from pymoo.operators.selection.tournament_selection import compare, TournamentSelection
from pymoo.util.display import MultiObjectiveDisplay
from pymoo.util.dominator import Dominator
from pymoo.util.misc import find_duplicates, has_feasible
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.util.randomized_argsort import randomized_argsort


# ---------------------------------------------------------------------------------------------------------
# Binary Tournament Selection Function
# ---------------------------------------------------------------------------------------------------------


def binary_tournament(pop, P, algorithm, **kwargs):
    if P.shape[1] != 2:
        raise ValueError("Only implemented for binary tournament!")

    tournament_type = algorithm.tournament_type
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):

        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop[a].CV > 0.0 or pop[b].CV > 0.0:
            S[i] = compare(a, pop[a].CV, b, pop[b].CV, method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible
        else:

            if tournament_type == 'comp_by_dom_and_crowding':
                rel = Dominator.get_relation(pop[a].F, pop[b].F)
                if rel == 1:
                    S[i] = a
                elif rel == -1:
                    S[i] = b

            elif tournament_type == 'comp_by_rank_and_crowding':
                S[i] = compare(a, pop[a].rank, b, pop[b].rank,
                               method='smaller_is_better')

            else:
                raise Exception("Unknown tournament type.")

            # if rank or domination relation didn't make a decision compare by crowding
            if np.isnan(S[i]):
                S[i] = compare(a, pop[a].get("crowding"), b, pop[b].get("crowding"),
                               method='larger_is_better', return_random_if_equal=True)

    return S[:, None].astype(np.int, copy=False)


# =========================================================================================================
# Implementation
# =========================================================================================================


class NSGA2(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 selection=TournamentSelection(func_comp=binary_tournament),
                 crossover=SimulatedBinaryCrossover(eta=15, prob=0.9),
                 mutation=PolynomialMutation(prob=None, eta=20),
                 eliminate_duplicates=True,
                 n_offsprings=None,
                 display=MultiObjectiveDisplay(),
                 **kwargs):
        """

        Parameters
        ----------
        pop_size : {pop_size}
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}
        n_offsprings : {n_offsprings}

        """

        kwargs['individual'] = Individual(rank=np.inf, crowding=-1)
        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=RankAndCrowdingSurvival(),
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=n_offsprings,
                         display=display,
                         **kwargs)

        self.tournament_type = 'comp_by_dom_and_crowding'

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.pop[self.pop.get("rank") == 0]


# ---------------------------------------------------------------------------------------------------------
# Survival Selection
# ---------------------------------------------------------------------------------------------------------


class RankAndCrowdingSurvival(Survival):

    def __init__(self) -> None:
        super().__init__(filter_infeasible=True)

    def _do(self, problem, pop, n_survive, D=None, **kwargs):

        # get the objective space values and objects
        F = pop.get("F").astype(np.float, copy=False)

        # the final indices of surviving individuals
        survivors = []

        # do the non-dominated sorting until splitting front
        fronts = NonDominatedSorting().do(F, n_stop_if_ranked=n_survive)

        for k, front in enumerate(fronts):

            # calculate the crowding distance of the front
            crowding_of_front = calc_crowding_distance(F[front, :])

            # save rank and crowding in the individual class
            for j, i in enumerate(front):
                pop[i].set("rank", k)
                pop[i].set("crowding", crowding_of_front[j])

            # current front sorted by crowding distance if splitting
            if len(survivors) + len(front) > n_survive:
                I = randomized_argsort(crowding_of_front, order='descending', method='numpy')
                I = I[:(n_survive - len(survivors))]

            # otherwise take the whole front unsorted
            else:
                I = np.arange(len(front))

            # extend the survivors by all or selected individuals
            survivors.extend(front[I])

        return pop[survivors]


def calc_crowding_distance(F, filter_out_duplicates=True):
    n_points, n_obj = F.shape

    if n_points <= 2:
        return np.full(n_points, np.inf)

    else:

        if filter_out_duplicates:
            # filter out solutions which are duplicates - duplicates get a zero finally
            is_unique = np.where(np.logical_not(find_duplicates(F, epsilon=1e-24)))[0]
        else:
            # set every point to be unique without checking it
            is_unique = np.arange(n_points)

        # index the unique points of the array
        _F = F[is_unique]

        # sort each column and get index
        I = np.argsort(_F, axis=0, kind='mergesort')

        # sort the objective space values for the whole matrix
        _F = _F[I, np.arange(n_obj)]

        # calculate the distance from each point to the last and next
        dist = np.row_stack([_F, np.full(n_obj, np.inf)]) - np.row_stack([np.full(n_obj, -np.inf), _F])

        # calculate the norm for each objective - set to NaN if all values are equal
        norm = np.max(_F, axis=0) - np.min(_F, axis=0)
        norm[norm == 0] = np.nan

        # prepare the distance to last and next vectors
        dist_to_last, dist_to_next = dist, np.copy(dist)
        dist_to_last, dist_to_next = dist_to_last[:-1] / norm, dist_to_next[1:] / norm

        # if we divide by zero because all values in one columns are equal replace by none
        dist_to_last[np.isnan(dist_to_last)] = 0.0
        dist_to_next[np.isnan(dist_to_next)] = 0.0

        # sum up the distance to next and last and norm by objectives - also reorder from sorted list
        J = np.argsort(I, axis=0)
        _cd = np.sum(dist_to_last[J, np.arange(n_obj)] + dist_to_next[J, np.arange(n_obj)], axis=1) / n_obj

        # save the final vector which sets the crowding distance for duplicates to zero to be eliminated
        crowding = np.zeros(n_points)
        crowding[is_unique] = _cd

    # crowding[np.isinf(crowding)] = 1e+14
    return crowding


parse_doc_string(NSGA2.__init__)



import numpy as np

from pymoo.algorithms.so_nelder_mead import max_expansion_factor
from pymoo.model.crossover import Crossover


class NelderMeadCrossover(Crossover):

    def __init__(self, problem, **kwargs):
        super().__init__(problem.n_var + 1, 1, **kwargs)

    def do(self, problem, pop, parents, **kwargs):
        n = problem.n_var - 1

        # obtain the matrices from the population
        X = pop.get("X")[parents]
        F = pop.get("F")[parents][..., 0]

        # the array to fill
        _X = np.full((len(X), problem.n_var), np.nan)

        for k, P in enumerate(parents):
            # sort the x values by their corresponding F values
            x = X[k, :][np.argsort(F[k])]

            # calculate the centroid of n best points
            centroid = x[:n + 1].mean(axis=0)

            # calculate the vector from the worst to the centroid
            v = centroid - x[n + 1]

            # maximum factor until the boundaries are hit
            max_factor = max_expansion_factor(centroid, v, problem)

            # randomly chose the extension, expansion or contraction through a factor
            factor = np.random.random() * min(3, max_factor)

            # calculate the offspring
            _X[k] = x[n + 1] + factor * v

        return pop.new("X", _X)



# START load_data
import numpy as np

np.random.seed(3)

ideal_point = np.array([0.15, 0.1, 0.2, 0.1, 0.1])
nadir_point = np.array([0.85, 0.9, 0.95, 0.9, 0.85])

F = np.random.random((1, 5)) * (nadir_point - ideal_point) + ideal_point
print(F)
# END load_data


# START radar
from pymoo.visualization.radar import Radar

plot = Radar(bounds=[ideal_point, nadir_point], normalize_each_objective=False)
plot.add(F)
plot.show()
# END radar


# START radar_norm
plot = Radar(bounds=[ideal_point, nadir_point])
plot.add(F)
plot.show()
# END radar_norm

# START radar_custom
F = np.random.random((6, 5)) * (nadir_point - ideal_point) + ideal_point

plot = Radar(bounds=[ideal_point, nadir_point],
             axis_style={"color": 'blue'},
             point_style={"color": 'red', 's': 30})
plot.add(F[:3], color="red", alpha=0.8)
plot.add(F[3:], color="green", alpha=0.8)
plot.show()
# END radar_custom



from pymoo.model.crossover import Crossover


class NoCrossover(Crossover):

    def __init__(self):
        super().__init__(1, 1, 0.0)

    def do(self, problem, pop, parents, **kwargs):
        return pop.new("X", pop.get("X"))



import numpy as np

from pymoo.performance_indicator.gd import GD
from pymoo.performance_indicator.igd import IGD
from pymoo.performance_indicator.hv import Hypervolume
from pymoo.util.termination.f_tol import MultiObjectiveSpaceToleranceTermination


def pareto_front_if_possible(problem):
    try:
        return problem.pareto_front()
    except:
        return None


class Output:

    def __init__(self, default_width=12) -> None:
        super().__init__()
        self.default_width = default_width
        self.attrs = []

    def append(self, name, number, format_if_float=True, width=None):
        if width is None:
            width = self.default_width
        if format_if_float and isinstance(number, float):
            number = self.format_float(number, width)

        self.attrs.append([name, number, width])

    def format_float(self, number, width):
        if number >= 10 or number * 1e5 < 1:
            return f"%.{width - 7}E" % number
        else:
            return f"%.{width - 3}f" % number

    def extend(self, *args):
        for arg in args:
            self.append(*arg)

    def clear(self):
        self.attrs = []

    def create_regex(self):
        return " | ".join(["{}"] * len(self.attrs))

    def do(self):
        if len(self.attrs) > 0:
            regex = self.create_regex()
            val = regex.format(*[str(val).rjust(width) for _, val, width in self.attrs])
            print(val)

    def header(self):
        regex = self.create_regex()
        s = regex.format(*[str(name).center(width) for name, _, width in self.attrs])
        print("=" * len(s))
        print(s)
        print("=" * len(s))


class Display:

    def __init__(self, output=None, attributes=None):
        super().__init__()
        self.output = output
        if self.output is None:
            self.output = Output()

        self.display_header = True
        self.pareto_front_is_available = None
        self.pf = None
        self.attributes = attributes

    def do(self, problem, evaluator, algorithm, pf=None):

        try:

            # if this runs for the first time - executed only once!
            if self.pareto_front_is_available is None:

                if isinstance(pf, bool) and not pf:
                    self.pf, self.pareto_front_is_available = None, False

                # try to get the pareto front from the problem
                elif pf is None or (isinstance(pf, bool) and pf):
                    self.pf = pareto_front_if_possible(problem)
                    self.pareto_front_is_available = self.pf is not None

                # the pf should be given directly
                else:
                    self.pf = pf
                    self.pf, self.pareto_front_is_available = pf, True

            self.output.clear()

            # get the actual attributes to display
            self._do(problem, evaluator, algorithm)

            if self.attributes is not None:
                self.output.attrs = [attr for attr in self.output.attrs if attr[0] in self.attributes]

            if self.display_header:
                self.output.header()

            # print the actually line
            self.output.do()

        # catch any exception to make sure the algorithm does not fail because of printing
        except:
            print("WARNING: Error while preparing the output to be printed.")
            self._do(problem, evaluator, algorithm)

        self.display_header = False

    def _do(self, problem, evaluator, algorithm):
        self.output.extend(*[('n_gen', algorithm.n_gen, False, 5), ('n_eval', evaluator.n_eval, False, 7)])


class SingleObjectiveDisplay(Display):

    def __init__(self, favg=True, **kwargs):
        super().__init__(**kwargs)
        self.favg = favg

    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)

        opt = algorithm.opt[0]
        F, CV, feasible = algorithm.pop.get("F", "CV", "feasible")
        feasible = np.where(feasible[:, 0])[0]

        if problem.n_constr > 0:
            self.output.append("cv (min)", opt.CV[0])
            self.output.append("cv (avg)", np.mean(CV))

        if len(feasible) > 0:
            _F = F[feasible]
            self.output.append("fopt", opt.F[0])
            if self.favg:
                self.output.append("favg", np.mean(_F))
        else:
            self.output.append("fopt", "-")
            if self.favg:
                self.output.append("favg", "-")


class MultiObjectiveDisplay(Display):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.term = MultiObjectiveSpaceToleranceTermination()

    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)

        F, CV, feasible = algorithm.pop.get("F", "CV", "feasible")
        feasible = np.where(feasible[:, 0])[0]

        if problem.n_constr > 0:
            self.output.append("cv (min)", CV.min())
            self.output.append("cv (avg)", np.mean(CV))

        if self.pareto_front_is_available:
            igd, gd, hv = "-", "-", "-"
            if len(feasible) > 0:
                _F = algorithm.opt.get("F")
                igd, gd = IGD(self.pf).calc(_F), GD(self.pf).calc(_F)
                if problem.n_obj == 2:
                    hv = Hypervolume(pf=self.pf).calc(_F)

            self.output.extend(*[('igd', igd), ('gd', gd)])
            if problem.n_obj == 2:
                self.output.append("hv", hv)

        else:
            self.output.append("n_nds", len(algorithm.opt), width=7)
            self.term.do_continue(algorithm)

            delta_ideal, delta_nadir, delta_f, hist_delta_max = "-", "-", "-", "-"

            if len(self.term.metrics) > 0:
                metric = self.term.metrics[-1]
                delta_ideal = metric["delta_ideal"]
                delta_nadir = metric["delta_nadir"]
                delta_f = metric["delta_f"]

            self.output.append("delta_ideal", delta_ideal)
            self.output.append("delta_nadir", delta_nadir)
            self.output.append("delta_f", delta_f)



class Termination:

    def __init__(self) -> None:
        """
        Base class for the implementation of a termination criterion for an algorithm.
        """
        super().__init__()

        # the algorithm can be forced to terminate by setting this attribute to true
        self.force_termination = False

    def do_continue(self, algorithm):
        """

        Whenever the algorithm objects wants to know whether it should continue or not it simply
        asks the termination criterion for it.

        Parameters
        ----------
        algorithm : class
            The algorithm object that is asking if it has terminated or not.

        Returns
        -------
        do_continue : bool
            Whether the algorithm has terminated or not.

        """

        if self.force_termination:
            return False
        else:
            return self._do_continue(algorithm)

    # the concrete implementation of the algorithm
    def _do_continue(self, algorithm, **kwargs):
        pass

    def has_terminated(self, algorithm):
        """
        Instead of asking if the algorithm should continue it can also ask if it has terminated.
        (just negates the continue method.)
        """
        return not self.do_continue(algorithm)



import numpy as np

from pymoo.model.population import Population


class Crossover:
    """
    The crossover combines parents to offsprings. Some crossover are problem specific and use additional information.
    This class must be inherited from to provide a crossover method to an algorithm.
    """

    def __init__(self, n_parents, n_offsprings, prob=0.9):
        self.prob = prob
        self.n_parents = n_parents
        self.n_offsprings = n_offsprings

    def do(self, problem, pop, parents, **kwargs):
        """

        This method executes the crossover on the parents. This class wraps the implementation of the class
        that implements the crossover.

        Parameters
        ----------
        problem: class
            The problem to be solved. Provides information such as lower and upper bounds or feasibility
            conditions for custom crossovers.

        pop : Population
            The population as an object

        parents: numpy.array
            The select parents of the population for the crossover

        kwargs : dict
            Any additional data that might be necessary to perform the crossover. E.g. constants of an algorithm.

        Returns
        -------
        offsprings : Population
            The off as a matrix. n_children rows and the number of columns is equal to the variable
            length of the problem.

        """

        if self.n_parents != parents.shape[1]:
            raise ValueError('Exception during crossover: Number of parents differs from defined at crossover.')

        # get the design space matrix form the population and parents
        X = pop.get("X")[parents.T].copy()

        # now apply the crossover probability
        do_crossover = np.random.random(len(parents)) < self.prob

        # execute the crossover
        _X = self._do(problem, X, **kwargs)

        X[:, do_crossover, :] = _X[:, do_crossover, :]

        # flatten the array to become a 2d-array
        X = X.reshape(-1, X.shape[-1])

        # create a population object
        off = pop.new("X", X)

        return off



import numpy as np

from pymoo.model.crossover import Crossover
from pymoo.operators.crossover.util import crossover_mask


class ExponentialCrossover(Crossover):

    def __init__(self, prob_exp=0.75, **kwargs):
        super().__init__(2, 2, **kwargs)
        self.prob_exp = prob_exp

    def _do(self, problem, X, **kwargs):

        # get the X of parents and count the matings
        _, n_matings, n_var = X.shape

        # the mask do to the crossover
        M = np.full((n_matings, n_var), False)

        # start point of crossover
        n = np.random.randint(0, n_var, size=X.shape[1])

        # the probabilities are calculated beforehand
        r = np.random.random((n_matings, n_var)) < self.prob_exp

        # create for each individual the crossover range
        for i in range(n_matings):

            # the actual index where we start
            start = n[i]
            for j in range(problem.n_var):

                # the current position where we are pointing to
                current = (start + j) % problem.n_var

                # replace only if random value keeps being smaller than CR
                if r[i, current]:
                    M[i, current] = True
                else:
                    break

        _X = crossover_mask(X, M)
        return _X



from pymoo.algorithms.nsga2 import NSGA2
from pymoo.factory import get_problem
from pymoo.optimize import minimize
from pymoo.visualization.scatter import Scatter

problem = get_problem("zdt1")

algorithm = NSGA2(pop_size=100)

res = minimize(problem,
               algorithm,
               ('n_gen', 200),
               seed=1,
               verbose=False)

plot = Scatter()
plot.add(problem.pareto_front(), plot_type="line", color="black", alpha=0.7)
plot.add(res.F, color="red")
plot.show()



from abc import abstractmethod


class Mutation:

    def __init__(self) -> None:
        super().__init__()
        self.algorithm = None
        self.problem = None

    def do(self, problem, pop, **kwargs):
        """

        Mutate variables in a genetic way.

        Parameters
        ----------
        problem : class
            The problem instance - specific information such as variable bounds might be needed.
        pop : Population
            A population object

        Returns
        -------
        Y : Population
            The mutated population.

        """

        Y = self._do(problem, pop.get("X"), **kwargs)
        return pop.new("X", Y)

    def _do(self, problem, X, **kwargs):
        pass



import numpy as np

from pymoo.model.evaluator import Evaluator
from pymoo.model.individual import Individual
from pymoo.model.population import Population


class GradientApproximation:

    def __init__(self, problem, epsilon=None, evaluator=None) -> None:
        super().__init__()
        self.epsilon = epsilon
        if self.epsilon is None:
            self.epsilon = np.sqrt(np.finfo(float).eps)
        self.problem = problem
        self.evaluator = evaluator if evaluator is not None else Evaluator()

    def do(self, individual, **kwargs):
        prob = self.problem
        n_var, n_obj, n_constr = prob.n_var, prob.n_obj, prob.n_constr
        individual = Individual(X=individual.X)
        self.evaluator.eval(self.problem, Population.create(individual))

        dF = np.zeros((n_obj, n_var))
        dG = np.zeros((n_constr, n_var))

        for i in range(n_var):
            x = np.copy(individual.X)
            x[i] += self.epsilon

            eps_F, _, eps_G = self.evaluator.eval(prob, x)
            dF[:, i] = (eps_F - individual.F) / self.epsilon

            if eps_G is not None:
                dG[:, i] = (eps_G - individual.G) / self.epsilon

        return dF, dG


def approx_jacobian(x, func, epsilon, *args):
    """
    Approximate the Jacobian matrix of a callable function.

    Parameters
    ----------
    x : array_like
        The state vector at which to compute the Jacobian matrix.
    func : callable f(x,*args)
        The vector-valued function.
    epsilon : float
        The perturbation used to determine the partial derivatives.
    args : sequence
        Additional arguments passed to func.

    Returns
    -------
    An array of dimensions ``(lenf, lenx)`` where ``lenf`` is the length
    of the outputs of `func`, and ``lenx`` is the number of elements in
    `x`.

    Notes
    -----
    The approximation is done using forward differences.

    """



class InfillCriterion:

    def __init__(self,
                 repair=None,
                 eliminate_duplicates=None,
                 n_max_iterations=100,
                 **kwargs):

        super().__init__()
        self.n_max_iterations = n_max_iterations
        self.eliminate_duplicates = eliminate_duplicates
        self.repair = repair

    def do(self, problem, pop, n_offsprings, **kwargs):

        # the population object to be used
        off = pop.new()

        # infill counter - counts how often the mating needs to be done to fill up n_offsprings
        n_infills = 0

        # iterate until enough offsprings are created
        while len(off) < n_offsprings:

            # how many offsprings are remaining to be created
            n_remaining = n_offsprings - len(off)

            # do the mating
            _off = self._do(problem, pop, n_remaining, **kwargs)

            # repair the individuals if necessary
            if self.repair:
                _off = self.repair.do(problem, _off, **kwargs)

            if self.eliminate_duplicates is not None:
                _off = self.eliminate_duplicates.do(_off, pop, off)

            # if more offsprings than necessary - truncate them randomly
            if len(off) + len(_off) > n_offsprings:
                # IMPORTANT: Interestingly, this makes a difference in performance
                n_remaining = n_offsprings - len(off)
                _off = _off[:n_remaining]

            # add to the offsprings and increase the mating counter
            off = off.merge(_off)
            n_infills += 1

            # if no new offsprings can be generated within a pre-specified number of generations
            if n_infills > self.n_max_iterations:
                break

        return off

    def _do(self, problem, pop, n_offsprings, **kwargs):
        pass



import numpy as np

from pymoo.model.repair import Repair
from pymoo.util.misc import at_least_2d_array


def bounds_back(problem, X):
    only_1d = (X.ndim == 1)
    X = at_least_2d_array(X)

    if problem.xl is not None and problem.xu is not None:
        xl = np.repeat(problem.xl[None, :], X.shape[0], axis=0)
        xu = np.repeat(problem.xu[None, :], X.shape[0], axis=0)

        # otherwise bounds back into the feasible space
        _range = xu - xl
        X[X < xl] = (xl + np.mod((xl - X), _range))[X < xl]
        X[X > xu] = (xu - np.mod((X - xu), _range))[X > xu]

    if only_1d:
        return X[0, :]
    else:
        return X


class BoundsBackRepair(Repair):

    def _do(self, problem, pop, **kwargs):
        # bring back to bounds if violated through crossover - bounce back strategy
        X = pop.get("X")
        xl = np.repeat(problem.xl[None, :], X.shape[0], axis=0)
        xu = np.repeat(problem.xu[None, :], X.shape[0], axis=0)

        # otherwise bounds back into the feasible space
        _range = xu - xl
        X[X < xl] = (xl + np.mod((xl - X), _range))[X < xl]
        X[X > xu] = (xu - np.mod((X - xu), _range))[X > xu]

        return pop.new("X", X)



from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm


class MemeticAlgorithm(GeneticAlgorithm):

    def _next(self):
        # do the mating using the current population
        self.off = self._mating(self.pop)

        ################################################################
        # Add a local optimization here
        ################################################################

        # evaluate the offspring
        self.evaluator.eval(self.problem, self.off, algorithm=self)

        # merge the offsprings with the current population
        self.pop = self.pop.merge(self.off)

        # the do survival selection
        self.pop = self.survival.do(self.problem, self.pop, self.pop_size, algorithm=self)



from abc import abstractmethod

import numpy as np


class Survival:

    def __init__(self, filter_infeasible=True) -> None:
        """
        The survival process is implemented inheriting from this class, which selects from a population only
        specific individuals to survive. This base class can take care of splitting the feasible and infeasible
        solutions first. By default infeasible solutions are simply sorted by their constraint violation.


        Parameters
        ----------

        filter_infeasible : bool
            Whether for the survival infeasible solutions should be filtered first

        """
        super().__init__()
        self.filter_infeasible = filter_infeasible

    def do(self, problem, pop, n_survive, return_indices=False, **kwargs):

        # if the split should be done beforehand
        if self.filter_infeasible and problem.n_constr > 0:
            feasible, infeasible = split_by_feasibility(pop, sort_infeasbible_by_cv=True)

            # if there was no feasible solution was added at all
            if len(feasible) == 0:
                survivors = pop[infeasible[:n_survive]]

            # if there are feasible solutions in the population
            else:
                survivors = pop.new()

                # if feasible solution do exist
                if len(feasible) > 0:
                    survivors = self._do(problem, pop[feasible], min(len(feasible), n_survive), **kwargs)

                # if infeasible solutions needs to be added
                if len(survivors) < n_survive:
                    least_infeasible = infeasible[:n_survive - len(feasible)]
                    survivors = survivors.merge(pop[least_infeasible])

        else:
            survivors = self._do(problem, pop, n_survive, **kwargs)

        if return_indices:
            H = {}
            for k, ind in enumerate(pop):
                H[ind] = k
            return [H[survivor] for survivor in survivors]
        else:
            return survivors

    @abstractmethod
    def _do(self, problem, pop, n_survive, **kwargs):
        pass


def split_by_feasibility(pop, sort_infeasbible_by_cv=True):
    CV = pop.get("CV")

    b = (CV <= 0)

    feasible = np.where(b)[0]
    infeasible = np.where(np.logical_not(b))[0]

    if sort_infeasbible_by_cv:
        infeasible = infeasible[np.argsort(CV[infeasible, 0])]

    return feasible, infeasible



import autograd.numpy as anp
from autograd.core import VJPNode, vspace, backward_pass
from autograd.tracer import new_box, isbox, toposort


# runs the function by making sure the calculations are traced using autograd
def run_and_trace(fun, x, *args, **kwargs):
    start_node = VJPNode.new_root()

    start_box = new_box(x, 0, start_node)
    out = fun(start_box, *args, **kwargs)

    return start_box, out



def calc_jacobian(start, end):

    # if the end_box is not a box - autograd can not track back
    if not isbox(end):
        return vspace(start.shape).zeros()

    # the final jacobian matrices
    jac = []

    # the backward pass is done for each objective function once
    for j in range(end.shape[1]):
        b = anp.zeros(end.shape)
        b[:, j] = 1
        n = new_box(b, 0, VJPNode.new_root())
        _jac = backward_pass(n, end._node)
        jac.append(_jac)

    jac = anp.stack(jac, axis=1)

    return jac




# START perm_prob
import numpy as np
from pymoo.model.problem import Problem


class MyProblem(Problem):

    def __init__(self, my_list):
        self.correct = np.argsort(my_list)
        super().__init__(n_var=len(my_list), n_obj=1, n_constr=0, xl=0, xu=1, elementwise_evaluation=True)

    def _evaluate(self, x, out, *args, **kwargs):
        pheno = np.argsort(x)
        out["F"] = - float((self.correct == pheno).sum())
        out["pheno"] = pheno
        out["hash"] = hash(str(pheno))


# END perm_prob


# START dupl
from pymoo.model.duplicate import ElementwiseDuplicateElimination


class MyElementwiseDuplicateElimination(ElementwiseDuplicateElimination):

    def is_equal(self, a, b):
        return a.get("hash")[0] == b.get("hash")[0]


# END dupl


# START problem
np.random.seed(2)
list_to_sort = np.random.random(20)
problem = MyProblem(list_to_sort)
print("Sorted by", np.argsort(list_to_sort))
# END problem


# START solve
from pymoo.algorithms.so_brkga import BRKGA
from pymoo.optimize import minimize

algorithm = BRKGA(
    n_elites=200,
    n_offsprings=700,
    n_mutants=100,
    bias=0.7,
    eliminate_duplicates=MyElementwiseDuplicateElimination())

res = minimize(problem,
               algorithm,
               ("n_gen", 75),
               seed=1,
               verbose=False)

print("Best solution found: \nX = %s\nF = %s" % (res.X, res.F))
print("Solution", res.opt.get("pheno")[0])
print("Optimum ", np.argsort(list_to_sort))
# END solve



from pymoo.model.termination import Termination


class MaximumFunctionCallTermination(Termination):

    def __init__(self, n_max_evals) -> None:
        super().__init__()
        self.n_max_evals = n_max_evals

        if self.n_max_evals is None:
            self.n_max_evals = float("inf")

    def _do_continue(self, algorithm):
        return algorithm.evaluator.n_eval < self.n_max_evals




from pymoo.algorithms.nsga2 import NSGA2
from pymoo.factory import get_problem
from pymoo.optimize import minimize
from pymoo.util.display import Display
import numpy as np


class MyDisplay(Display):

    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)
        self.output.append("metric_a", np.mean(algorithm.pop.get("X")))
        self.output.append("metric_b", np.mean(algorithm.pop.get("F")))


problem = get_problem("zdt2")

algorithm = NSGA2(pop_size=100)

res = minimize(problem,
               algorithm,
               ('n_gen', 10),
               seed=1,
               display=MyDisplay(),
               verbose=True)




import copy
import distutils
import os
import sys
import traceback

import setuptools
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

from pymoo.version import __version__

# ---------------------------------------------------------------------------------------------------------
# GENERAL
# ---------------------------------------------------------------------------------------------------------


__name__ = "pymoo"
__author__ = "Julian Blank"
__url__ = "https://pymoo.org"

data = dict(
    name=__name__,
    version=__version__,
    author=__author__,
    url=__url__,
    python_requires='>=3.6',
    author_email="blankjul@egr.msu.edu",
    description="Multi-Objective Optimization in Python",
    license='Apache License 2.0',
    keywords="optimization",
    install_requires=['numpy>=1.15', 'scipy>=1.1', 'matplotlib>=3', 'autograd>=1.3', 'cma==2.7'],
    platforms='any',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics'
    ]
)


# ---------------------------------------------------------------------------------------------------------
# OTHER METADATA
# ---------------------------------------------------------------------------------------------------------


# update the readme.rst to be part of setup
def readme():
    with open('README.rst') as f:
        return f.read()


def packages():
    return ["pymoo"] + ["pymoo." + e for e in setuptools.find_packages(where='pymoo')]


data['long_description'] = readme()
data['long_description_content_type'] = 'text/x-rst'
data['packages'] = packages()


# ---------------------------------------------------------------------------------------------------------
# Extensions
# ---------------------------------------------------------------------------------------------------------


# ============================================================
# MacOSX FIX for compiling modules
# ============================================================

def is_new_osx():
    name = distutils.util.get_platform()
    if sys.platform != "darwin":
        return False
    elif name.startswith("macosx-10"):
        minor_version = int(name.split("-")[1].split(".")[1])
        if minor_version >= 7:
            return True
        else:
            return False
    else:
        return False


# fix compiling for new macosx!
if is_new_osx():
    os.environ['CFLAGS'] = '-stdlib=libc++'


# ============================================================
# Module for Compilation - Throws an Exception if Failing
# ============================================================


# exception that is thrown when the build fails
class CompilingFailed(Exception):
    pass


# try to compile, if not possible throw exception
def construct_build_ext(build_ext):
    class WrappedBuildExt(build_ext):
        def run(self):
            try:
                build_ext.run(self)
            except BaseException as e:
                raise CompilingFailed(e)

        def build_extension(self, ext):
            try:
                build_ext.build_extension(self, ext)
            except BaseException as e:
                raise CompilingFailed(e)

    return WrappedBuildExt


# ============================================================
# SETUP
# ============================================================


def run_setup(setup_args):
    # try to add compilation to the setup - if fails just do default
    try:

        do_cythonize = False
        if "--cythonize" in sys.argv:
            do_cythonize = True
            sys.argv.remove("--cythonize")

        # copy the kwargs for compiling purpose - if it fails setup_args remain unchanged
        kwargs = copy.deepcopy(setup_args)
        kwargs['cmdclass'] = {}

        try:
            import numpy as np
            kwargs['include_dirs'] = [np.get_include()]
        except BaseException:
            raise CompilingFailed(
                "NumPy libraries must be installed for compiled extensions! Speedups are not enabled.")

        # return the object for building which allows installation with no compilation
        kwargs['cmdclass']['build_ext'] = construct_build_ext(build_ext)

        # all the modules must be finally added here
        kwargs['ext_modules'] = []
        cython_folder = os.path.join("pymoo", "cython")
        cython_files = os.listdir(cython_folder)

        # if the pyx files should be translated and then compiled
        if do_cythonize:
            from Cython.Build import cythonize
            kwargs['ext_modules'] = cythonize("pymoo/cython/*.pyx")

        # otherwise use the existing pyx files - normal case during pip installation
        else:

            # find all cpp files in czthon folder
            cpp_files = [f for f in cython_files if f.endswith(".cpp")]

            # add for each file an extension object to be compiled
            for source in cpp_files:
                ext = Extension("pymoo.cython.%s" % source[:-4], [os.path.join(cython_folder, source)])
                kwargs['ext_modules'].append(ext)

        if len(kwargs['ext_modules']) == 0:
            print('*' * 75)
            print("WARNING: No modules for compilation available. To compile pyx files, execute:")
            print("make compile-with-cython")
            print('*' * 75)
            raise CompilingFailed()

        setup(**kwargs)
        print('*' * 75)
        print("Compilation Successful.")
        print("Installation with Compilation succeeded.")
        print('*' * 75)

    except BaseException:

        # retrieve the original input arguments and execute default setup
        kwargs = setup_args
        setup(**kwargs)

        # get information why compiling has failed
        ex_type, ex_value, ex_traceback = sys.exc_info()

        print('*' * 75)
        print("WARNING: Compilation Failed.")
        print("WARNING:", ex_type)
        print("WARNING:", ex_value)
        print()
        print("=" * 75)
        traceback.print_exc()
        print("=" * 75)
        print()
        print("WARNING: For the compiled libraries numpy is required. Please make sure they are installed")
        print("WARNING: pip install numpy")
        print("WARNING: Also, make sure you have a compiler for C++!")

        print('*' * 75)
        print("Plain Python installation succeeded.")
        print('*' * 75)


run_setup(data)




from pymoo.model.termination import Termination


class TerminationCollection(Termination):

    def __init__(self, *args) -> None:
        super().__init__()
        self.terminations = args

    def _do_continue(self, algorithm):
        for term in self.terminations:
            if not term.do_continue(algorithm):
                return False
        return True



from pymoo.algorithms.so_cmaes import CMAES
from pymoo.factory import get_problem
from pymoo.optimize import minimize

problem = get_problem("sphere")

algorithm = CMAES()

res = minimize(problem,
               algorithm,
               seed=1,
               verbose=False)

print(f"Best solution found: \nX = {res.X}\nF = {res.X}\nCV= {res.CV}")


from pymoo.algorithms.so_genetic_algorithm import GA
from pymoo.optimize import minimize
from pymoo.problems.single.mopta08 import MOPTA08

problem = MOPTA08("/Users/blankjul/Downloads/mopta_fortran_unix/exec")
algorithm = GA(n_offsprings=20, return_least_infeasible=True)

res = minimize(problem,
               algorithm,
               verbose=True
               )

print(res.F, res.F)



import numpy as np
from scipy.spatial.distance import cdist

from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm
from pymoo.factory import get_decomposition
from pymoo.operators.crossover.simulated_binary_crossover import SimulatedBinaryCrossover
from pymoo.operators.mutation.polynomial_mutation import PolynomialMutation
from pymoo.operators.sampling.random_sampling import FloatRandomSampling
from pymoo.util.display import MultiObjectiveDisplay
from pymoo.util.misc import set_if_none


# =========================================================================================================
# Implementation
# =========================================================================================================

class MOEAD(GeneticAlgorithm):

    def __init__(self,
                 ref_dirs,
                 n_neighbors=20,
                 decomposition='auto',
                 prob_neighbor_mating=0.9,
                 display=MultiObjectiveDisplay(),
                 **kwargs):
        """

        Parameters
        ----------
        ref_dirs
        n_neighbors
        decomposition
        prob_neighbor_mating
        display
        kwargs
        """

        self.n_neighbors = n_neighbors
        self.prob_neighbor_mating = prob_neighbor_mating
        self.decomposition = decomposition

        set_if_none(kwargs, 'pop_size', len(ref_dirs))
        set_if_none(kwargs, 'sampling', FloatRandomSampling())
        set_if_none(kwargs, 'crossover', SimulatedBinaryCrossover(prob=1.0, eta=20))
        set_if_none(kwargs, 'mutation', PolynomialMutation(prob=None, eta=20))
        set_if_none(kwargs, 'survival', None)
        set_if_none(kwargs, 'selection', None)

        super().__init__(display=display, **kwargs)

        # initialized when problem is known
        self.ref_dirs = ref_dirs

        if self.ref_dirs.shape[0] < self.n_neighbors:
            print("Setting number of neighbours to population size: %s" % self.ref_dirs.shape[0])
            self.n_neighbors = self.ref_dirs.shape[0]

        # neighbours includes the entry by itself intentionally for the survival method
        self.neighbors = np.argsort(cdist(self.ref_dirs, self.ref_dirs), axis=1, kind='quicksort')[:, :self.n_neighbors]

    def _initialize(self):

        if isinstance(self.decomposition, str):

            # set a string
            decomp = self.decomposition

            # for one or two objectives use tchebi otherwise pbi
            if decomp == 'auto':
                if self.problem.n_obj <= 2:
                    decomp = 'tchebi'
                else:
                    decomp = 'pbi'

            # set the decomposition object
            self._decomposition = get_decomposition(decomp)

        else:
            self._decomposition = self.decomposition

        super()._initialize()
        self.ideal_point = np.min(self.pop.get("F"), axis=0)

    def _next(self):
        repair, crossover, mutation = self.mating.repair, self.mating.crossover, self.mating.mutation

        # retrieve the current population
        pop = self.pop

        # iterate for each member of the population in random order
        for i in np.random.permutation(len(pop)):

            # all neighbors of this individual and corresponding weights
            N = self.neighbors[i, :]

            if np.random.random() < self.prob_neighbor_mating:
                parents = N[np.random.permutation(self.n_neighbors)][:crossover.n_parents]
            else:
                parents = np.random.permutation(self.pop_size)[:crossover.n_parents]

            # do recombination and create an offspring
            off = crossover.do(self.problem, pop, parents[None, :])
            off = mutation.do(self.problem, off)
            off = off[np.random.randint(0, len(off))]

            # repair first in case it is necessary
            if repair:
                off = self.repair.do(self.problem, off, algorithm=self)

            # evaluate the offspring
            self.evaluator.eval(self.problem, off)

            # update the ideal point
            self.ideal_point = np.min(np.vstack([self.ideal_point, off.F]), axis=0)

            # calculate the decomposed values for each neighbor
            FV = self._decomposition.do(pop[N].get("F"), weights=self.ref_dirs[N, :], ideal_point=self.ideal_point)
            off_FV = self._decomposition.do(off.F[None, :], weights=self.ref_dirs[N, :], ideal_point=self.ideal_point)

            # get the absolute index in F where offspring is better than the current F (decomposed space)
            I = np.where(off_FV < FV)[0]
            pop[N[I]] = off


# parse_doc_string(MOEAD.__init__)



from pymoo.algorithms.nsga2 import NSGA2
from pymoo.factory import get_problem, get_termination
from pymoo.optimize import minimize

problem = get_problem("zdt3")
algorithm = NSGA2(pop_size=100)
termination = get_termination("n_gen", 10)

res = minimize(problem,
               algorithm,
               termination,
               pf=problem.pareto_front(),
               seed=1,
               verbose=True)


import numpy as np

from pymoo.model.crossover import Crossover
from pymoo.model.mutation import Mutation
from pymoo.model.sampling import Sampling


def prepare_processing(mask, operators):
    process = []

    # create a numpy array of mask if it is not yet
    mask = np.array(mask)

    for val in np.unique(mask):

        # check if operator for that type was defined
        if val not in operators:
            raise Exception("Operator for type %s was not defined." % val)

        # append it as a processing type
        process.append({
            "type": val,
            "mask": mask == val,
            "operator": operators[val]
        })

    return process


def apply_mixed_variable_operation(problem, process, fun):

    # the result to be returned
    ret = []

    # save the original bounds of the problem
    _n_var, _xl, _xu = problem.n_var, problem.xl, problem.xu

    # iterate through all the different operators that should be applied
    for entry in process:
        # get the mask and the operator
        mask, operator = entry["mask"], entry["operator"]

        # copy the arrays of the problem and cast them to float
        problem.n_var, problem.xl, problem.xu = mask.sum(), _xl[mask], _xu[mask]

        # perform the crossover
        ret.append(fun(mask, operator))

    # reset the original bounds of the problem
    problem.n_var = _n_var
    problem.xl = _xl
    problem.xu = _xu

    return ret


def concatenate_mixed_variables(problem, process, ret):
    # find the minimum of returned individuals and make them equal among operators
    n_rows = min([len(e) for e in ret])
    ret = [e[:n_rows] for e in ret]

    # create the result array and set the values for each operator
    X = np.full((n_rows, problem.n_var), np.nan, dtype=np.object)

    for i in range(len(process)):
        mask, _X = process[i]["mask"], ret[i]
        X[:, mask] = _X

    return X


class MixedVariableCrossover(Crossover):

    def __init__(self, mask, operators):

        n_parents = np.unique(np.array([op.n_parents for op in operators.values()]))
        if len(n_parents) > 1:
            raise Exception("All crossovers need to have the same number of parents!")

        n_offsprings = np.unique(np.array([op.n_offsprings for op in operators.values()]))
        if len(n_offsprings) > 1:
            raise Exception("All crossovers need to have the same number of offsprings!")

        super().__init__(n_parents[0], n_offsprings[0])
        self.process = prepare_processing(mask, operators)

    def _do(self, problem, X, **kwargs):

        _, n_matings, n_var = X.shape

        def fun(mask, operator):
            return operator._do(problem, X[..., mask], **kwargs)

        ret = apply_mixed_variable_operation(problem, self.process, fun)

        # for the crossover the concatenation is different through the 3d arrays.
        X = np.full((self.n_offsprings, n_matings, n_var), np.nan, dtype=np.object)
        for i in range(len(self.process)):
            mask, _X = self.process[i]["mask"], ret[i]
            X[..., mask] = _X

        return X


class MixedVariableMutation(Mutation):

    def __init__(self, mask, operators):
        super().__init__()
        self.process = prepare_processing(mask, operators)

    def _do(self, problem, X, **kwargs):
        def fun(mask, operator):
            return operator._do(problem, X[:, mask], **kwargs)

        ret = apply_mixed_variable_operation(problem, self.process, fun)
        return concatenate_mixed_variables(problem, self.process, ret)


class MixedVariableSampling(Sampling):

    def __init__(self, mask, operators):
        super().__init__()
        self.process = prepare_processing(mask, operators)

    def _do(self, problem, n_samples, **kwargs):
        def fun(mask, operator):
            return operator._do(problem, n_samples, **kwargs)

        ret = apply_mixed_variable_operation(problem, self.process, fun)
        return concatenate_mixed_variables(problem, self.process, ret)



import numpy as np

from pymoo.model.repair import Repair
from pymoo.util.misc import at_least_2d_array


def repair_out_of_bounds_manually(X, xl, xu):

    only_1d = (X.ndim == 1)
    X = at_least_2d_array(X)

    if xl is not None:
        xl = np.repeat(xl[None, :], X.shape[0], axis=0)
        X[X < xl] = xl[X < xl]

    if xu is not None:
        xu = np.repeat(xu[None, :], X.shape[0], axis=0)
        X[X > xu] = xu[X > xu]

    if only_1d:
        return X[0, :]
    else:
        return X


def repair_out_of_bounds(problem, X):
    return repair_out_of_bounds_manually(X, problem.xl, problem.xu)


class OutOfBoundsRepair(Repair):

    def _do(self, problem, pop, **kwargs):
        X = pop.get("X")
        pop.set("X", repair_out_of_bounds(problem, X))
        return pop



import numpy as np

from pymoo.model.crossover import Crossover
from pymoo.operators.crossover.util import crossover_mask


class BiasedCrossover(Crossover):

    def __init__(self, bias, **kwargs):
        super().__init__(2, 1, **kwargs)
        self.bias = bias

    def _do(self, problem, X, **kwargs):
        _, n_matings, n_var = X.shape
        M = np.random.random((n_matings, n_var)) < self.bias
        _X = crossover_mask(X, M)
        return _X

    def do(self, problem, pop, parents, **kwargs):
        off = super().do(problem, pop, parents, **kwargs)
        return off[:int(len(off) / 2)]



import numpy as np

from pymoo.model.decision_making import DecisionMaking, normalize, NeighborFinder, find_outliers_upper_tail


class HighTradeoffPointsInverted(DecisionMaking):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def _do(self, F, **kwargs):

        n, m = F.shape

        if self.normalize:
            F = normalize(F, self.ideal_point, self.nadir_point, estimate_bounds_if_none=True)

        neighbors_finder = NeighborFinder(F, n_min_neigbors="auto")

        mu = np.full(n, - np.inf)

        # for each solution in the set calculate the least amount of improvement per unit deterioration
        for i in range(n):

            # neighbors to the current point
            neighbors = neighbors_finder.find(i)

            # calculate the trade-off to all neighbours
            diff = F[neighbors] - F[i]

            # calculate sacrifice and gain
            sacrifice = np.maximum(0, diff).sum(axis=1)
            gain = np.maximum(0, -diff).sum(axis=1)

            np.warnings.filterwarnings('ignore')
            tradeoff = sacrifice / gain

            # otherwise find the one with the smalled one
            mu[i] = np.nanmean(tradeoff)

        return find_outliers_upper_tail(mu)

if __name__ == '__main__':

    F = np.array([[2, 13], [3, 11], [7, 9], [9, 4], [12, 3]])
    #_, w = PseudoWeights(np.array([0.5, 0.5])).do(F, return_pseudo_weights=True)
    HighTradeoffPointsInverted(normalize=False).do(F)



import autograd.numpy as anp

from pymoo.model.problem import Problem


# always derive from the main problem for the evaluation
class MyProblem(Problem):

    def __init__(self, const_1=5, const_2=0.1):

        # define lower and upper bounds -  1d array with length equal to number of variable
        xl = -5 * anp.ones(10)
        xu = 5 * anp.ones(10)

        super().__init__(n_var=10, n_obj=1, n_constr=2, xl=xl, xu=xu, evaluation_of="auto")

        # store custom variables needed for evaluation
        self.const_1 = const_1
        self.const_2 = const_2

    # implemented the function evaluation function - the arrays to fill are provided directly
    def _evaluate(self, x, out, *args, **kwargs):
        # define an objective function to be evaluated using var1
        f = anp.sum(anp.power(x, 2) - self.const_1 * anp.cos(2 * anp.pi * x), axis=1)

        # !!! only if a constraint value is positive it is violated !!!
        # set the constraint that x1 + x2 > var2
        g1 = (x[:, 0] + x[:, 1]) - self.const_2

        # set the constraint that x3 + x4 < var2
        g2 = self.const_2 - (x[:, 2] + x[:, 3])

        out["F"] = f
        out["G"] = anp.column_stack([g1, g2])


problem = MyProblem()
F, G, CV, feasible, dF, dG = problem.evaluate(anp.random.rand(100, 10),
                                              return_values_of=["F", "G", "CV", "feasible", "dF", "dG"])



from pymoo.model.algorithm import Algorithm
from pymoo.model.population import pop_from_array_or_individual
from pymoo.operators.sampling.latin_hypercube_sampling import LatinHypercubeSampling


class LocalSearch(Algorithm):

    def __init__(self,
                 x0=None,
                 sampling=LatinHypercubeSampling(),
                 n_sample_points="auto",
                 **kwargs):

        super().__init__(**kwargs)
        self.x0 = x0
        self.sampling = sampling
        self.n_sample_points = n_sample_points

    def initialize(self, problem, **kwargs):
        super().initialize(problem, **kwargs)

        if self.n_sample_points == "auto":
            self.n_sample_points = self.problem.n_var * 5

    def _initialize(self, **kwargs):
        super()._initialize(**kwargs)

        # no initial point is provided - sample in bounds and take the best
        if self.x0 is None:
            if not self.problem.has_bounds():
                raise Exception("Either provide an x0 or a problem with variable bounds!")

            self.pop = self.sampling.do(self.problem, self.n_sample_points)
        else:
            self.pop = pop_from_array_or_individual(self.x0)

        self.evaluator.eval(self.problem, self.pop, algorithm=self)
        self._set_optimum()
        self.x0 = self.opt[0]





# --------------------------------------------------------------------------------------------
# Single
# --------------------------------------------------------------------------------------------

# START ackley
import numpy as np

from pymoo.factory import get_problem
from pymoo.util.plotting import plot_problem_surface

problem = get_problem("ackley", n_var=2, a=20, b=1/5, c=2 * np.pi)
plot_problem_surface(problem, 100, plot_type="wireframe+contour")
# END ackley

# START rastrigin
from pymoo.factory import get_problem
from pymoo.util.plotting import plot_problem_surface

problem = get_problem("rastrigin", n_var=2)
plot_problem_surface(problem, 100, plot_type="wireframe+contour")
# END rastrigin

# START sphere
from pymoo.factory import get_problem
from pymoo.util.plotting import plot_problem_surface

problem = get_problem("sphere", n_var=2)
plot_problem_surface(problem, 100, plot_type="wireframe+contour")
# END sphere

# START zakharov
from pymoo.factory import get_problem
from pymoo.util.plotting import plot_problem_surface

problem = get_problem("zakharov", n_var=2)
plot_problem_surface(problem, 100, plot_type="wireframe+contour")
# END zakharov



# --------------------------------------------------------------------------------------------
# Multi
# --------------------------------------------------------------------------------------------


# START zdt1
from pymoo.factory import get_problem
from pymoo.util.plotting import plot

problem = get_problem("zdt1")
plot(problem.pareto_front(), no_fill=True)
# END zdt1

# START zdt2
from pymoo.factory import get_problem
from pymoo.util.plotting import plot

problem = get_problem("zdt2")
plot(problem.pareto_front(), no_fill=True)
# END zdt2

# START zdt3
from pymoo.factory import get_problem
from pymoo.util.plotting import plot

problem = get_problem("zdt3")
plot(problem.pareto_front(), no_fill=True)
# END zdt3

# START zdt4
from pymoo.factory import get_problem
from pymoo.util.plotting import plot

problem = get_problem("zdt4")
plot(problem.pareto_front(), no_fill=True)
# END zdt4

# START zdt5_no_norm
from pymoo.factory import get_problem
from pymoo.util.plotting import plot

problem = get_problem("zdt5", normalize=False)
plot(problem.pareto_front(), no_fill=True)
# END zdt5_no_norm

# START zdt5
from pymoo.factory import get_problem
from pymoo.util.plotting import plot

problem = get_problem("zdt5")
plot(problem.pareto_front(), no_fill=True)
# END zdt5

# START zdt6
from pymoo.factory import get_problem
from pymoo.util.plotting import plot

problem = get_problem("zdt6")
plot(problem.pareto_front(), no_fill=True)
# END zdt6

# START bnh
from pymoo.factory import get_problem
from pymoo.util.plotting import plot

problem = get_problem("bnh")
plot(problem.pareto_front(), no_fill=True)
# END bnh


# START rosenbrock
from pymoo.factory import get_problem
from pymoo.util.plotting import plot_problem_surface

problem = get_problem("rosenbrock", n_var=2)
plot_problem_surface(problem, 100, plot_type="wireframe+contour")
# END rosenbrock

# START griewank
from pymoo.factory import get_problem
from pymoo.util.plotting import plot_problem_surface

problem = get_problem("griewank", n_var=2)
plot_problem_surface(problem, 100, plot_type="wireframe+contour")
# END griewank

# START truss2d
from pymoo.factory import get_problem
from pymoo.visualization.scatter import Scatter

pf = get_problem("truss2d").pareto_front()

plot = Scatter(title="Pareto-front")
plot.add(pf, s=80, facecolors='none', edgecolors='r')
plot.add(pf, plot_type="line", color="black", linewidth=2)
plot.show()
# END truss2d

# START truss2d_log
plot.reset()
plot.do()
plot.apply(lambda ax: ax.set_yscale("log"))
plot.apply(lambda ax: ax.set_xscale("log"))
plot.show()
# END truss2d_log




# --------------------------------------------------------------------------------------------


# START from_string

from pymoo.factory import get_problem

p = get_problem("dtlz1_-1", n_var=20, n_obj=5)

# create a simple test problem from string
p = get_problem("Ackley")

# the input name is not case sensitive
p = get_problem("ackley")

# also input parameter can be provided directly
p = get_problem("dtlz1_-1", n_var=20, n_obj=5)

# END from_string


# --------------------------------------------------------------------------------------------

# START from_function

import numpy as np

# this will be the evaluation function that is called each time
from pymoo.model.problem import get_problem_from_func


def my_evaluate_func(x, out, *args, **kwargs):

    # define the objective as x^2
    f1 = np.sum(np.square(x - 2), axis=1)
    f2 = np.sum(np.square(x + 2), axis=1)
    out["F"] = np.column_stack([f1, f2])

    # x^2 < 2 constraint
    out["G"] = np.sum(np.square(x - 1), axis=1)


# load the problem from a function - define 3 variables with the same lower bound
problem = get_problem_from_func(my_evaluate_func, xl=-10, xu=10, n_var=3)
F, CV = problem.evaluate(np.random.rand(100, 3))

# or define a problem with varying lower and upper bounds
problem = get_problem_from_func(my_evaluate_func, xl=np.array([-10, -5, -10]), xu=np.array([10, 5, 10]))
F, CV = problem.evaluate(np.random.rand(100, 3))

# END from_string


from pymoo.algorithms.so_genetic_algorithm import GA
from pymoo.factory import get_problem
from pymoo.optimize import minimize
from pymoo.util.termination.f_tol_single import SingleObjectiveSpaceToleranceTermination

problem = get_problem("rastrigin")
algorithm = GA(pop_size=100)
termination = SingleObjectiveSpaceToleranceTermination()

res = minimize(problem,
               algorithm,
               termination,
               pf=problem.pareto_front(),
               seed=1,
               verbose=True)

print(res.opt.get("F"))
print(res.algorithm.n_gen)



import autograd.numpy as anp

from pymoo.model.problem import Problem


class Rastrigin(Problem):
    def __init__(self, n_var=2, A=10.0):
        super().__init__(n_var=n_var, n_obj=1, n_constr=0, xl=-5, xu=5, type_var=anp.double)
        self.A = A

    def _evaluate(self, x, out, *args, **kwargs):
        z = anp.power(x, 2) - self.A * anp.cos(2 * anp.pi * x)
        out["F"] = self.A * self.n_var + anp.sum(z, axis=1)

    def _calc_pareto_front(self):
        return 0

    def _calc_pareto_set(self):
        return anp.full(self.n_var, 0)



from pymoo.factory import get_problem, get_reference_directions

pf = get_problem("tnk").pareto_front()
pf = get_problem("osy").pareto_front()

# for other single the number of non-dominated points can be defined
pf = get_problem("zdt1").pareto_front(n_pareto_points=100)

# for DTLZ for example the reference direction should be provided, because the pareto front for the
# specific problem will depend on the factory for the reference lines
ref_dirs = get_reference_directions("das-dennis", 3, n_partitions=12)
pf = get_problem("dtlz1", n_var=7, n_obj=3).pareto_front(ref_dirs)



import numpy as np

from pymoo.algorithms.so_genetic_algorithm import GA
from pymoo.factory import get_problem
from pymoo.optimize import minimize

problem = get_problem("sphere")

X = np.random.random((500, problem.n_var))

algorithm = GA(sampling=X)

res = minimize(problem,
               algorithm,
               seed=1,
               verbose=False)

print("Best solution found: \nX = %s\nF = %s" % (res.X, res.F))


from pymoo.model.problem import Problem
from pymoo.vendor.go_benchmark_functions import *


class GlobalOptimizationProblem(Problem):

    def __init__(self, n_var=None, clazz=None, **kwargs):
        if clazz is None:
            raise Exception("Implementation must be provided.")

        self.object = clazz()

        if n_var is not None:
            self.object.change_dimensions(n_var)

        super().__init__(n_var=self.object.N,
                         n_obj=1,
                         n_constr=0,
                         xl=self.object.xmin,
                         xu=self.object.xmax,
                         elementwise_evaluation=True,
                         **kwargs)

    def _calc_pareto_set(self, *args, **kwargs):
        return np.array(self.object.global_optimum)

    def _calc_pareto_front(self, *args, **kwargs):
        return self.object.fglob

    def _evaluate(self, x, out, *args, **kwargs):
        F = np.array([self.object.fun(x)])
        F[np.isnan(F)] = np.inf
        out["F"] = F

    def success(self, x, **kwargs):
        return self.object.success(x, **kwargs)


def get_global_optimization_problem_options():
    PROBLEMS = [
        ("go-amgm", GlobalOptimizationProblem, {"clazz": AMGM}),
        ("go-ackley01", GlobalOptimizationProblem, {"clazz": Ackley01}),
        ("go-ackley02", GlobalOptimizationProblem, {"clazz": Ackley02}),
        ("go-ackley03", GlobalOptimizationProblem, {"clazz": Ackley03}),
        ("go-adjiman", GlobalOptimizationProblem, {"clazz": Adjiman}),
        ("go-alpine01", GlobalOptimizationProblem, {"clazz": Alpine01}),
        ("go-alpine02", GlobalOptimizationProblem, {"clazz": Alpine02}),
        ("go-bartelsconn", GlobalOptimizationProblem, {"clazz": BartelsConn}),
        ("go-beale", GlobalOptimizationProblem, {"clazz": Beale}),
        ("go-biggsexp02", GlobalOptimizationProblem, {"clazz": BiggsExp02}),
        ("go-biggsexp03", GlobalOptimizationProblem, {"clazz": BiggsExp03}),
        ("go-biggsexp04", GlobalOptimizationProblem, {"clazz": BiggsExp04}),
        ("go-biggsexp05", GlobalOptimizationProblem, {"clazz": BiggsExp05}),
        ("go-bird", GlobalOptimizationProblem, {"clazz": Bird}),
        ("go-bohachevsky1", GlobalOptimizationProblem, {"clazz": Bohachevsky1}),
        ("go-bohachevsky2", GlobalOptimizationProblem, {"clazz": Bohachevsky2}),
        ("go-bohachevsky3", GlobalOptimizationProblem, {"clazz": Bohachevsky3}),
        ("go-boxbetts", GlobalOptimizationProblem, {"clazz": BoxBetts}),
        ("go-branin01", GlobalOptimizationProblem, {"clazz": Branin01}),
        ("go-branin02", GlobalOptimizationProblem, {"clazz": Branin02}),
        ("go-brent", GlobalOptimizationProblem, {"clazz": Brent}),
        ("go-brown", GlobalOptimizationProblem, {"clazz": Brown}),
        ("go-bukin02", GlobalOptimizationProblem, {"clazz": Bukin02}),
        ("go-bukin04", GlobalOptimizationProblem, {"clazz": Bukin04}),
        ("go-bukin06", GlobalOptimizationProblem, {"clazz": Bukin06}),
        ("go-carromtable", GlobalOptimizationProblem, {"clazz": CarromTable}),
        ("go-chichinadze", GlobalOptimizationProblem, {"clazz": Chichinadze}),
        ("go-cigar", GlobalOptimizationProblem, {"clazz": Cigar}),
        ("go-cola", GlobalOptimizationProblem, {"clazz": Cola}),
        ("go-colville", GlobalOptimizationProblem, {"clazz": Colville}),
        ("go-corana", GlobalOptimizationProblem, {"clazz": Corana}),
        ("go-cosinemixture", GlobalOptimizationProblem, {"clazz": CosineMixture}),
        ("go-crossintray", GlobalOptimizationProblem, {"clazz": CrossInTray}),
        ("go-crosslegtable", GlobalOptimizationProblem, {"clazz": CrossLegTable}),
        ("go-crownedcross", GlobalOptimizationProblem, {"clazz": CrownedCross}),
        ("go-csendes", GlobalOptimizationProblem, {"clazz": Csendes}),
        ("go-cube", GlobalOptimizationProblem, {"clazz": Cube}),
        ("go-damavandi", GlobalOptimizationProblem, {"clazz": Damavandi}),
        ("go-devilliersglasser01", GlobalOptimizationProblem, {"clazz": DeVilliersGlasser01}),
        ("go-devilliersglasser02", GlobalOptimizationProblem, {"clazz": DeVilliersGlasser02}),
        ("go-deb01", GlobalOptimizationProblem, {"clazz": Deb01}),
        ("go-deb03", GlobalOptimizationProblem, {"clazz": Deb03}),
        ("go-decanomial", GlobalOptimizationProblem, {"clazz": Decanomial}),
        ("go-deceptive", GlobalOptimizationProblem, {"clazz": Deceptive}),
        ("go-deckkersaarts", GlobalOptimizationProblem, {"clazz": DeckkersAarts}),
        ("go-deflectedcorrugatedspring", GlobalOptimizationProblem, {"clazz": DeflectedCorrugatedSpring}),
        ("go-dixonprice", GlobalOptimizationProblem, {"clazz": DixonPrice}),
        ("go-dolan", GlobalOptimizationProblem, {"clazz": Dolan}),
        ("go-dropwave", GlobalOptimizationProblem, {"clazz": DropWave}),
        ("go-easom", GlobalOptimizationProblem, {"clazz": Easom}),
        ("go-eckerle4", GlobalOptimizationProblem, {"clazz": Eckerle4}),
        ("go-eggcrate", GlobalOptimizationProblem, {"clazz": EggCrate}),
        ("go-eggholder", GlobalOptimizationProblem, {"clazz": EggHolder}),
        ("go-elattarvidyasagardutta", GlobalOptimizationProblem, {"clazz": ElAttarVidyasagarDutta}),
        ("go-exp2", GlobalOptimizationProblem, {"clazz": Exp2}),
        ("go-exponential", GlobalOptimizationProblem, {"clazz": Exponential}),
        ("go-freudensteinroth", GlobalOptimizationProblem, {"clazz": FreudensteinRoth}),
        ("go-gear", GlobalOptimizationProblem, {"clazz": Gear}),
        ("go-giunta", GlobalOptimizationProblem, {"clazz": Giunta}),
        ("go-goldsteinprice", GlobalOptimizationProblem, {"clazz": GoldsteinPrice}),
        ("go-griewank", GlobalOptimizationProblem, {"clazz": Griewank}),
        ("go-gulf", GlobalOptimizationProblem, {"clazz": Gulf}),
        ("go-hansen", GlobalOptimizationProblem, {"clazz": Hansen}),
        ("go-hartmann3", GlobalOptimizationProblem, {"clazz": Hartmann3}),
        ("go-hartmann6", GlobalOptimizationProblem, {"clazz": Hartmann6}),
        ("go-helicalvalley", GlobalOptimizationProblem, {"clazz": HelicalValley}),
        ("go-himmelblau", GlobalOptimizationProblem, {"clazz": HimmelBlau}),
        ("go-holdertable", GlobalOptimizationProblem, {"clazz": HolderTable}),
        ("go-hosaki", GlobalOptimizationProblem, {"clazz": Hosaki}),
        ("go-infinity", GlobalOptimizationProblem, {"clazz": Infinity}),
        ("go-jennrichsampson", GlobalOptimizationProblem, {"clazz": JennrichSampson}),
        ("go-judge", GlobalOptimizationProblem, {"clazz": Judge}),
        ("go-katsuura", GlobalOptimizationProblem, {"clazz": Katsuura}),
        ("go-keane", GlobalOptimizationProblem, {"clazz": Keane}),
        ("go-kowalik", GlobalOptimizationProblem, {"clazz": Kowalik}),
        ("go-langermann", GlobalOptimizationProblem, {"clazz": Langermann}),
        ("go-lennardjones", GlobalOptimizationProblem, {"clazz": LennardJones}),
        ("go-leon", GlobalOptimizationProblem, {"clazz": Leon}),
        ("go-levy03", GlobalOptimizationProblem, {"clazz": Levy03}),
        ("go-levy05", GlobalOptimizationProblem, {"clazz": Levy05}),
        ("go-levy13", GlobalOptimizationProblem, {"clazz": Levy13}),
        ("go-matyas", GlobalOptimizationProblem, {"clazz": Matyas}),
        ("go-mccormick", GlobalOptimizationProblem, {"clazz": McCormick}),
        ("go-meyer", GlobalOptimizationProblem, {"clazz": Meyer}),
        ("go-michalewicz", GlobalOptimizationProblem, {"clazz": Michalewicz}),
        ("go-mielecantrell", GlobalOptimizationProblem, {"clazz": MieleCantrell}),
        ("go-mishra01", GlobalOptimizationProblem, {"clazz": Mishra01}),
        ("go-mishra02", GlobalOptimizationProblem, {"clazz": Mishra02}),
        ("go-mishra03", GlobalOptimizationProblem, {"clazz": Mishra03}),
        ("go-mishra04", GlobalOptimizationProblem, {"clazz": Mishra04}),
        ("go-mishra05", GlobalOptimizationProblem, {"clazz": Mishra05}),
        ("go-mishra06", GlobalOptimizationProblem, {"clazz": Mishra06}),
        ("go-mishra07", GlobalOptimizationProblem, {"clazz": Mishra07}),
        ("go-mishra08", GlobalOptimizationProblem, {"clazz": Mishra08}),
        ("go-mishra09", GlobalOptimizationProblem, {"clazz": Mishra09}),
        ("go-mishra10", GlobalOptimizationProblem, {"clazz": Mishra10}),
        ("go-mishra11", GlobalOptimizationProblem, {"clazz": Mishra11}),
        ("go-multimodal", GlobalOptimizationProblem, {"clazz": MultiModal}),
        ("go-needleeye", GlobalOptimizationProblem, {"clazz": NeedleEye}),
        ("go-newfunction01", GlobalOptimizationProblem, {"clazz": NewFunction01}),
        ("go-newfunction02", GlobalOptimizationProblem, {"clazz": NewFunction02}),
        ("go-oddsquare", GlobalOptimizationProblem, {"clazz": OddSquare}),
        ("go-parsopoulos", GlobalOptimizationProblem, {"clazz": Parsopoulos}),
        ("go-pathological", GlobalOptimizationProblem, {"clazz": Pathological}),
        ("go-paviani", GlobalOptimizationProblem, {"clazz": Paviani}),
        ("go-penholder", GlobalOptimizationProblem, {"clazz": PenHolder}),
        ("go-penalty01", GlobalOptimizationProblem, {"clazz": Penalty01}),
        ("go-penalty02", GlobalOptimizationProblem, {"clazz": Penalty02}),
        ("go-permfunction01", GlobalOptimizationProblem, {"clazz": PermFunction01}),
        ("go-permfunction02", GlobalOptimizationProblem, {"clazz": PermFunction02}),
        ("go-pinter", GlobalOptimizationProblem, {"clazz": Pinter}),
        ("go-plateau", GlobalOptimizationProblem, {"clazz": Plateau}),
        ("go-powell", GlobalOptimizationProblem, {"clazz": Powell}),
        ("go-powersum", GlobalOptimizationProblem, {"clazz": PowerSum}),
        ("go-price01", GlobalOptimizationProblem, {"clazz": Price01}),
        ("go-price02", GlobalOptimizationProblem, {"clazz": Price02}),
        ("go-price03", GlobalOptimizationProblem, {"clazz": Price03}),
        ("go-price04", GlobalOptimizationProblem, {"clazz": Price04}),
        ("go-qing", GlobalOptimizationProblem, {"clazz": Qing}),
        ("go-quadratic", GlobalOptimizationProblem, {"clazz": Quadratic}),
        ("go-quintic", GlobalOptimizationProblem, {"clazz": Quintic}),
        ("go-rana", GlobalOptimizationProblem, {"clazz": Rana}),
        ("go-rastrigin", GlobalOptimizationProblem, {"clazz": Rastrigin}),
        ("go-ratkowsky01", GlobalOptimizationProblem, {"clazz": Ratkowsky01}),
        ("go-ratkowsky02", GlobalOptimizationProblem, {"clazz": Ratkowsky02}),
        ("go-ripple01", GlobalOptimizationProblem, {"clazz": Ripple01}),
        ("go-ripple25", GlobalOptimizationProblem, {"clazz": Ripple25}),
        ("go-rosenbrock", GlobalOptimizationProblem, {"clazz": Rosenbrock}),
        ("go-rosenbrockmodified", GlobalOptimizationProblem, {"clazz": RosenbrockModified}),
        ("go-rotatedellipse01", GlobalOptimizationProblem, {"clazz": RotatedEllipse01}),
        ("go-rotatedellipse02", GlobalOptimizationProblem, {"clazz": RotatedEllipse02}),
        ("go-salomon", GlobalOptimizationProblem, {"clazz": Salomon}),
        ("go-sargan", GlobalOptimizationProblem, {"clazz": Sargan}),
        ("go-schaffer01", GlobalOptimizationProblem, {"clazz": Schaffer01}),
        ("go-schaffer02", GlobalOptimizationProblem, {"clazz": Schaffer02}),
        ("go-schaffer03", GlobalOptimizationProblem, {"clazz": Schaffer03}),
        ("go-schaffer04", GlobalOptimizationProblem, {"clazz": Schaffer04}),
        ("go-schwefel01", GlobalOptimizationProblem, {"clazz": Schwefel01}),
        ("go-schwefel02", GlobalOptimizationProblem, {"clazz": Schwefel02}),
        ("go-schwefel04", GlobalOptimizationProblem, {"clazz": Schwefel04}),
        ("go-schwefel06", GlobalOptimizationProblem, {"clazz": Schwefel06}),
        ("go-schwefel20", GlobalOptimizationProblem, {"clazz": Schwefel20}),
        ("go-schwefel21", GlobalOptimizationProblem, {"clazz": Schwefel21}),
        ("go-schwefel22", GlobalOptimizationProblem, {"clazz": Schwefel22}),
        ("go-schwefel26", GlobalOptimizationProblem, {"clazz": Schwefel26}),
        ("go-schwefel36", GlobalOptimizationProblem, {"clazz": Schwefel36}),
        ("go-shekel05", GlobalOptimizationProblem, {"clazz": Shekel05}),
        ("go-shekel07", GlobalOptimizationProblem, {"clazz": Shekel07}),
        ("go-shekel10", GlobalOptimizationProblem, {"clazz": Shekel10}),
        ("go-shubert01", GlobalOptimizationProblem, {"clazz": Shubert01}),
        ("go-shubert03", GlobalOptimizationProblem, {"clazz": Shubert03}),
        ("go-shubert04", GlobalOptimizationProblem, {"clazz": Shubert04}),
        ("go-sineenvelope", GlobalOptimizationProblem, {"clazz": SineEnvelope}),
        ("go-sixhumpcamel", GlobalOptimizationProblem, {"clazz": SixHumpCamel}),
        ("go-sodp", GlobalOptimizationProblem, {"clazz": Sodp}),
        ("go-sphere", GlobalOptimizationProblem, {"clazz": Sphere}),
        ("go-step", GlobalOptimizationProblem, {"clazz": Step}),
        ("go-step2", GlobalOptimizationProblem, {"clazz": Step2}),
        ("go-stochastic", GlobalOptimizationProblem, {"clazz": Stochastic}),
        ("go-stretchedv", GlobalOptimizationProblem, {"clazz": StretchedV}),
        ("go-styblinskitang", GlobalOptimizationProblem, {"clazz": StyblinskiTang}),
        ("go-testtubeholder", GlobalOptimizationProblem, {"clazz": TestTubeHolder}),
        ("go-threehumpcamel", GlobalOptimizationProblem, {"clazz": ThreeHumpCamel}),
        ("go-thurber", GlobalOptimizationProblem, {"clazz": Thurber}),
        ("go-treccani", GlobalOptimizationProblem, {"clazz": Treccani}),
        ("go-trefethen", GlobalOptimizationProblem, {"clazz": Trefethen}),
        ("go-trid", GlobalOptimizationProblem, {"clazz": Trid}),
        ("go-trigonometric01", GlobalOptimizationProblem, {"clazz": Trigonometric01}),
        ("go-trigonometric02", GlobalOptimizationProblem, {"clazz": Trigonometric02}),
        ("go-tripod", GlobalOptimizationProblem, {"clazz": Tripod}),
        ("go-ursem01", GlobalOptimizationProblem, {"clazz": Ursem01}),
        ("go-ursem03", GlobalOptimizationProblem, {"clazz": Ursem03}),
        ("go-ursem04", GlobalOptimizationProblem, {"clazz": Ursem04}),
        ("go-ursemwaves", GlobalOptimizationProblem, {"clazz": UrsemWaves}),
        ("go-ventersobiezcczanskisobieski", GlobalOptimizationProblem, {"clazz": VenterSobiezcczanskiSobieski}),
        ("go-vincent", GlobalOptimizationProblem, {"clazz": Vincent}),
        ("go-watson", GlobalOptimizationProblem, {"clazz": Watson}),
        ("go-wavy", GlobalOptimizationProblem, {"clazz": Wavy}),
        ("go-wayburnseader01", GlobalOptimizationProblem, {"clazz": WayburnSeader01}),
        ("go-wayburnseader02", GlobalOptimizationProblem, {"clazz": WayburnSeader02}),
        ("go-weierstrass", GlobalOptimizationProblem, {"clazz": Weierstrass}),
        ("go-whitley", GlobalOptimizationProblem, {"clazz": Whitley}),
        ("go-wolfe", GlobalOptimizationProblem, {"clazz": Wolfe}),
        ("go-xinsheyang01", GlobalOptimizationProblem, {"clazz": XinSheYang01}),
        ("go-xinsheyang02", GlobalOptimizationProblem, {"clazz": XinSheYang02}),
        ("go-xinsheyang03", GlobalOptimizationProblem, {"clazz": XinSheYang03}),
        ("go-xinsheyang04", GlobalOptimizationProblem, {"clazz": XinSheYang04}),
        ("go-xor", GlobalOptimizationProblem, {"clazz": Xor}),
        ("go-yaoliu04", GlobalOptimizationProblem, {"clazz": YaoLiu04}),
        ("go-yaoliu09", GlobalOptimizationProblem, {"clazz": YaoLiu09}),
        ("go-zacharov", GlobalOptimizationProblem, {"clazz": Zacharov}),
        ("go-zerosum", GlobalOptimizationProblem, {"clazz": ZeroSum}),
        ("go-zettl", GlobalOptimizationProblem, {"clazz": Zettl}),
        ("go-zimmerman", GlobalOptimizationProblem, {"clazz": Zimmerman}),
        ("go-zirilli", GlobalOptimizationProblem, {"clazz": Zirilli})
    ]

    return PROBLEMS


if __name__ == "__main__":
    for key, clazz in dict(globals()).items():
        if isinstance(clazz, type) and issubclass(clazz, Benchmark):
            name = clazz.__name__
            print(f"\"go-{name.lower()}\",")
            # print(f"(\"go-{name.lower()}\", GlobalOptimizationProblem, {{\"clazz\":{name}}}),")
    print("done")



import numpy as np

from pymoo.model.indicator import Indicator
from pymoo.util.misc import vectorized_cdist


def euclidean_distance(a, b, norm=None):
    return np.sqrt((((a - b) / norm) ** 2).sum(axis=1))


def modified_distance(z, a, norm=None):
    d = a - z
    d[d < 0] = 0
    d = d / norm
    return np.sqrt((d ** 2).sum(axis=1))


class DistanceIndicator(Indicator):

    def __init__(self, pf, dist_func, axis, **kwargs):
        super().__init__(pf=pf, **kwargs)
        self.dist_func = dist_func
        self.axis = axis

    def _calc(self, F):
        D = vectorized_cdist(self.pf, F, func_dist=self.dist_func, norm=self.range)
        return np.mean(np.min(D, axis=self.axis))





import numpy as np
from pymoo.model.problem import Problem


class SubsetProblem(Problem):
    def __init__(self,
                 L,
                 n_max
                 ):
        super().__init__(n_var=len(L), n_obj=1, n_constr=1, elementwise_evaluation=True)
        self.L = L
        self.n_max = n_max

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = np.sum(self.L[x])
        out["G"] = (self.n_max - np.sum(x)) ** 2


# create the actual problem to be solved
np.random.seed(1)
L = np.array([np.random.randint(100) for _ in range(100)])
n_max = 10
problem = SubsetProblem(L, n_max)

from pymoo.model.crossover import Crossover
from pymoo.model.mutation import Mutation
from pymoo.model.sampling import Sampling


class MySampling(Sampling):

    def _do(self, problem, n_samples, **kwargs):
        X = np.full((n_samples, problem.n_var), False, dtype=np.bool)

        for k in range(n_samples):
            I = np.random.permutation(problem.n_var)[:problem.n_max]
            X[k, I] = True

        return X


class BinaryCrossover(Crossover):
    def __init__(self):
        super().__init__(2, 1)

    def _do(self, problem, X, **kwargs):
        n_parents, n_matings, n_var = X.shape

        _X = np.full((self.n_offsprings, n_matings, problem.n_var), False)

        for k in range(n_matings):
            p1, p2 = X[0, k], X[1, k]

            both_are_true = np.logical_and(p1, p2)
            _X[0, k, both_are_true] = True

            n_remaining = problem.n_max - np.sum(both_are_true)

            I = np.where(np.logical_xor(p1, p2))[0]

            S = I[np.random.permutation(len(I))][:n_remaining]
            _X[0, k, S] = True

        return _X


class MyMutation(Mutation):
    def _do(self, problem, X, **kwargs):
        for i in range(X.shape[0]):
            X[i, :] = X[i, :]
            is_false = np.where(np.logical_not(X[i, :]))[0]
            is_true = np.where(X[i, :])[0]
            X[i, np.random.choice(is_false)] = True
            X[i, np.random.choice(is_true)] = False

        return X


from pymoo.algorithms.so_genetic_algorithm import GA
from pymoo.optimize import minimize

algorithm = GA(
    pop_size=100,
    sampling=MySampling(),
    crossover=BinaryCrossover(),
    mutation=MyMutation(),
    eliminate_duplicates=True)

res = minimize(problem,
               algorithm,
               ('n_gen', 75),
               seed=1,
               verbose=True)

print("Function value: %s" % res.F[0])
print("Subset:", np.where(res.X)[0])

opt = np.sort(np.argsort(L)[:n_max])
print("Optimal Subset:", opt)
print("Optimal Function Value: %s" % L[opt].sum())



from pymoo.algorithms.so_genetic_algorithm import GA
from pymoo.factory import get_problem
from pymoo.optimize import minimize

problem = get_problem("g01")

algorithm = GA(
    pop_size=100,
    eliminate_duplicates=True)

res = minimize(problem,
               algorithm,
               termination=('n_gen', 50),
               seed=1,
               verbose=False)



from abc import abstractmethod


class Selection:

    def __init__(self) -> None:
        """
        This class is used to select parents for the mating or other evolutionary operators.
        Several strategies can be used to increase the selection pressure.
        """
        super().__init__()

    def do(self, pop, n_select, n_parents=2, **kwargs):
        """
        Choose from the population new individuals to be selected.

        Parameters
        ----------
        pop : :class:`~pymoo.model.population.Population`
            The population which should be selected from. Some criteria from the design or objective space
            might be used for the selection. Therefore, only the number of individual might be not enough.

        n_select : int
            Number of individuals to select.

        n_parents : int
            Number of parents needed to create an offspring.

        Returns
        -------
        I : numpy.array
            Indices of selected individuals.

        """

        return self._do(pop, n_select, n_parents, **kwargs)

    @abstractmethod
    def _do(self, pop, n_select, n_parents, **kwargs):
        pass


from datetime import datetime
from itertools import combinations

import numpy as np
import scipy
import scipy.spatial

from pymoo.model.population import Population
from pymoo.model.sampling import Sampling


def parameter_less(F, CV):
    ret = np.copy(F)
    parameter_less = np.max(F, axis=0) + CV

    infeasible = CV > 0
    ret[infeasible] = parameter_less[infeasible]

    return ret


def swap(M, a, b):
    tmp = M[a]
    M[a] = M[b]
    M[b] = tmp


# repairs a numpy array to be in bounds
def repair(X, xl, xu):
    larger_than_xu = X[0, :] > xu
    X[0, larger_than_xu] = xu[larger_than_xu]

    smaller_than_xl = X[0, :] < xl
    X[0, smaller_than_xl] = xl[smaller_than_xl]

    return X


def unique_rows(a):
    a = np.ascontiguousarray(a)
    unique_a = np.unique(a.view([('', a.dtype)] * a.shape[1]))
    return unique_a.view(a.dtype).reshape((unique_a.shape[0], a.shape[1]))


def parameter_less_constraints(F, CV, F_max=None):
    if F_max is None:
        F_max = np.max(F)
    has_constraint_violation = CV > 0
    F[has_constraint_violation] = CV[has_constraint_violation] + F_max
    return F


def random_permuations(n, l):
    perms = []
    for i in range(n):
        perms.append(np.random.permutation(l))
    P = np.concatenate(perms)
    return P


def get_duplicates(M):
    res = []
    I = np.lexsort([M[:, i] for i in reversed(range(0, M.shape[1]))])
    S = M[I, :]

    i = 0

    while i < S.shape[0] - 1:
        l = []
        while np.all(S[i, :] == S[i + 1, :]):
            l.append(I[i])
            i += 1
        if len(l) > 0:
            l.append(I[i])
            res.append(l)
        i += 1

    return res


def euclidean_distance(a, b):
    return np.sqrt(((a - b) ** 2).sum(axis=1))


def norm_euclidean_distance(problem):
    return lambda a, b: np.sqrt((((a - b) / (problem.xu - problem.xl)) ** 2).sum(axis=1))


def cdist(A, B, **kwargs):
    return scipy.spatial.distance.cdist(A, B, **kwargs)


def vectorized_cdist(A, B, func_dist=euclidean_distance, fill_diag_with_inf=False, **kwargs):
    u = np.repeat(A, B.shape[0], axis=0)
    v = np.tile(B, (A.shape[0], 1))

    D = func_dist(u, v, **kwargs)
    M = np.reshape(D, (A.shape[0], B.shape[0]))

    if fill_diag_with_inf:
        np.fill_diagonal(M, np.inf)

    return M


def norm_eucl_dist(problem, A, B, **kwargs):
    return vectorized_cdist(A, B, func_dist=norm_euclidean_distance(problem), **kwargs)


def covert_to_type(problem, X):
    if problem.type_var == np.double:
        return X.astype(np.double)
    elif problem.type_var == np.int:
        return np.round(X).astype(np.int)
    elif problem.type_var == np.bool:
        return X < (problem.xu - problem.xl) / 2


def find_duplicates(X, epsilon=1e-16):
    # calculate the distance matrix from each point to another
    D = cdist(X, X)

    # set the diagonal to infinity
    D[np.triu_indices(len(X))] = np.inf

    # set as duplicate if a point is really close to this one
    is_duplicate = np.any(D < epsilon, axis=1)

    return is_duplicate


def at_least_2d_array(x, extend_as="row"):
    if not isinstance(x, np.ndarray):
        x = np.array([x])

    if x.ndim == 1:
        if extend_as == "row":
            x = x[None, :]
        elif extend_as == "column":
            x = x[:, None]

    return x


def to_1d_array_if_possible(x):
    if not isinstance(x, np.ndarray):
        x = np.array([x])

    if x.ndim == 2:
        if x.shape[0] == 1 or x.shape[1] == 1:
            x = x.flatten()

    return x


def stack(*args, flatten=True):
    if not flatten:
        ps = np.concatenate([e[None, ...] for e in args])
    else:
        ps = np.row_stack(args)
    return ps


def all_combinations(A, B):
    u = np.repeat(A, B.shape[0], axis=0)
    v = np.tile(B, A.shape[0])
    return np.column_stack([u, v])


def pop_from_sampling(problem, sampling, n_initial_samples, pop=None):
    # the population type can be different - (different type of individuals)
    if pop is None:
        pop = Population()

    # provide a whole population object - (individuals might be already evaluated)
    if isinstance(sampling, Population):
        pop = sampling

    else:
        # if just an X array create a pop
        if isinstance(sampling, np.ndarray):
            pop = pop.new("X", sampling)

        elif isinstance(sampling, Sampling):
            # use the sampling
            pop = sampling.do(problem, n_initial_samples, pop=pop)

        else:
            return None

    return pop


def evaluate_if_not_done_yet(evaluator, problem, pop, algorithm=None):
    I = np.where(pop.get("F") == None)[0]
    if len(I) > 0:
        pop[I] = evaluator.eval(problem, pop[I], algorithm=algorithm)


def set_if_none(kwargs, str, val):
    if str not in kwargs:
        kwargs[str] = val


def set_if_none_from_tuples(kwargs, *args):
    for key, val in args:
        if key not in kwargs:
            kwargs[key] = val


def calc_perpendicular_distance(N, ref_dirs):
    u = np.tile(ref_dirs, (len(N), 1))
    v = np.repeat(N, len(ref_dirs), axis=0)

    norm_u = np.linalg.norm(u, axis=1)

    scalar_proj = np.sum(v * u, axis=1) / norm_u
    proj = scalar_proj[:, None] * u / norm_u[:, None]
    val = np.linalg.norm(proj - v, axis=1)
    matrix = np.reshape(val, (len(N), len(ref_dirs)))

    return matrix


def distance_of_closest_points_to_others(X):
    D = vectorized_cdist(X, X)
    np.fill_diagonal(D, np.inf)
    return D.argmin(axis=1), D.min(axis=1)


def time_to_int(t):
    td = datetime.strptime(t, '%H:%M:%S') - datetime(1900, 1, 1)
    return td.total_seconds()


def powerset(iterable):
    for n in range(len(iterable) + 1):
        yield from combinations(iterable, n)


def intersect(a, b):
    H = set()
    for entry in b:
        H.add(entry)

    ret = []
    for entry in a:
        if entry in H:
            ret.append(entry)

    return ret


def has_feasible(pop):
    return np.any(pop.get("feasible"))


def to_numpy(a):
    return np.array(a)



import os


class Configuration:
    show_compile_hint = True
    parse_custom_docs = False


# returns the directory to be used for imports
def get_pymoo():
    return os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
